# coding=utf-8
"""
This script is used to building training and validation datasets.
"""

from __future__ import print_function, absolute_import

import json
import tensorflow as tf
from collections import namedtuple
from functools import partial
from os import makedirs
from os.path import join, isdir
from .constants import SEED

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


tf.app.flags.DEFINE_string("binary_dir", "./binary",
                           """The directory for storing binary datasets.""")
tf.app.flags.DEFINE_string('dataset', 'C9H7N.PBE',
                           """Define the dataset to use. This is also the name
                           of the xyz file to load.""")
tf.app.flags.DEFINE_integer('k_max', 3,
                            """The maximum k under the many-body-expansion 
                            scheme.""")
tf.app.flags.DEFINE_boolean('include_all_k', True,
                            """Include all k-body terms from k = 1 to k_max.""")
tf.app.flags.DEFINE_boolean('forces', False,
                            """Set this to True to enable atomic forces.""")
tf.app.flags.DEFINE_integer('num_parallel_calls', 64,
                            """The number elements to process in parallel.""")

FLAGS = tf.app.flags.FLAGS


def get_filenames(train=True, dataset_name=None):
  """
  Return the binary data file and the config file.

  Args:
    train: boolean indicating whether the training data file or the testing file
      should be returned.
    dataset_name: a `str` as the name of the dataset.

  Returns:
    (binfile, jsonfile): the binary file to read data and the json file to read
      configs..

  """
  binary_dir = FLAGS.binary_dir
  if not isdir(binary_dir):
    makedirs(binary_dir)

  name = dataset_name or FLAGS.dataset
  records = {"train": (join(binary_dir, "{}-train.tfrecords".format(name)),
                       join(binary_dir, "{}-train.json".format(name))),
             "test": (join(binary_dir, "{}-test.tfrecords".format(name)),
                      join(binary_dir, "{}-test.json".format(name)))}

  if train:
    return records['train']
  else:
    return records['test']


"""
These namedtuples are just used to manage the decoded example.
"""
EnergyExample = namedtuple("EnergyExample", (
  "features",
  "energy",
  "occurs",
  "weights",
  "y_weight",
))

ForcesExample = namedtuple("ForcesExample", (
  "features",
  "energy",
  "occurs",
  "weights",
  "y_weight",
  "forces",
  "coef",
  "indexing"
))


def decode_protobuf(example_proto, cnk=None, ck2=None, num_atom_types=None,
                    atomic_forces=False, num_f_components=None,
                    num_entries=None):
  """
  Decode the protobuf into a tuple of tensors.

  Args:
    example_proto: A scalar string Tensor, a single serialized Example.
      See `_parse_single_example_raw` documentation for more details.
    cnk: an `int` as the value of C(N,k).
    ck2: an `int` as the value of C(k,2).
    num_atom_types: an `int` as the number of atom types.
    atomic_forces: a `bool` indicating whether atomic forces should be included
      or not.
    num_f_components: an `int` as the maximum number of force components. This
      must be set if `atomic_forces` is True.
    num_entries: an `int` as the number of entries per each force component.
      This must be set if `atomic_forces` is True.

  Returns:
    example: a decoded `TFExample` from the TFRecord file.

  """
  if not atomic_forces:
    example = tf.parse_single_example(
      example_proto,
      # Defaults are not specified since both keys are required.
      features={
        'features': tf.FixedLenFeature([], tf.string),
        'energy': tf.FixedLenFeature([], tf.string),
        'occurs': tf.FixedLenFeature([], tf.string),
        'weights': tf.FixedLenFeature([], tf.string),
        'loss_weight': tf.FixedLenFeature([], tf.float32)
      })
  else:
    example = tf.parse_single_example(
      example_proto,
      features={
        'features': tf.FixedLenFeature([], tf.string),
        'energy': tf.FixedLenFeature([], tf.string),
        'occurs': tf.FixedLenFeature([], tf.string),
        'weights': tf.FixedLenFeature([], tf.string),
        'loss_weight': tf.FixedLenFeature([], tf.float32),
        'coef': tf.FixedLenFeature([], tf.string),
        'indexing': tf.FixedLenFeature([], tf.string),
        'forces': tf.FixedLenFeature([], tf.string)
      })
    assert num_f_components > 0 and num_entries > 0

  features = tf.decode_raw(example['features'], tf.float32)
  features.set_shape([cnk * ck2])
  features = tf.reshape(features, [1, cnk, ck2])

  energy = tf.decode_raw(example['energy'], tf.float64)
  energy.set_shape([1])
  energy = tf.squeeze(energy)

  occurs = tf.decode_raw(example['occurs'], tf.float32)
  occurs.set_shape([num_atom_types])
  occurs = tf.reshape(occurs, [1, 1, num_atom_types])

  weights = tf.decode_raw(example['weights'], tf.float32)
  weights.set_shape([cnk, ])
  weights = tf.reshape(weights, [1, cnk, 1])

  y_weight = tf.cast(example['loss_weight'], tf.float32)

  if atomic_forces:
    coef = tf.decode_raw(example['coef'], tf.float32)
    coef.set_shape([cnk * ck2 * 6])
    coef = tf.reshape(coef, [cnk, ck2 * 6])

    indexing = tf.decode_raw(example['indexing'], tf.int32)
    indexing.set_shape([num_f_components * num_entries])
    indexing = tf.reshape(indexing, [num_f_components, num_entries])

    forces = tf.decode_raw(example['forces'], tf.float64)
    forces.set_shape([num_f_components])

    return ForcesExample(features=features,
                         energy=energy,
                         occurs=occurs,
                         weights=weights,
                         y_weight=y_weight,
                         forces=forces,
                         coef=coef,
                         indexing=indexing)

  else:
    return EnergyExample(features=features,
                         energy=energy,
                         occurs=occurs,
                         weights=weights,
                         y_weight=y_weight)


def get_configs(for_training=True, dataset_name=None):
  """
  Return the configs for inputs.

  Args:
    for_training: boolean indicating if one should return settings for training
      or validation.
    dataset_name: a `str` as the name of the dataset.

  Returns:
    configs: a `dict` of configs.

  """
  _, json_file = get_filenames(train=for_training, dataset_name=dataset_name)
  with open(json_file) as f:
    return dict(json.load(f))


def get_dataset_size(dataset_name, for_training=True):
  """
  Return the total number of examples in the given datatset.

  Args:
    dataset_name: a `str` as the name of the dataset.
    for_training: a `bool` indicating whether we should return the size of the
      training set or validation set.

  Returns:
    num_examples: an `int` as the number of examples in the dataset.

  """
  return len(get_configs(for_training, dataset_name)['lookup_indices'])


def next_batch(dataset_name, for_training=True, batch_size=50, num_epochs=None,
               shuffle=True):
  """
  Provide batched inputs for kCON.

  Args:
    dataset_name: a `str` as the name of the dataset.
    for_training: a `bool` selecting between the training (True) and validation
      (False) data.
    batch_size: an `int` as the number of examples per batch.
    num_epochs: an `int` as the maximum number of epochs to run.
    shuffle: a `bool` indicating whether the batches shall be shuffled or not.

  Returns:
    next_batch: a tuple of Tensors.

  """

  with tf.device('/cpu:0'):
    tfrecods_file, _ = get_filenames(
      train=for_training,
      dataset_name=dataset_name
    )

    configs = get_configs(for_training=for_training, dataset_name=dataset_name)
    shape = configs["shape"]
    cnk = shape[0]
    ck2 = shape[1]
    num_atom_types = configs["num_atom_types"]
    dataset_size = get_dataset_size(dataset_name, for_training=for_training)

    if FLAGS.forces:
      num_f_components, num_entries = configs["indexing_shape"]
    else:
      num_f_components, num_entries = None, None

    # Set the number of parallel calls (threads).
    num_parallel_calls = min(batch_size, FLAGS.num_parallel_calls)

    # Initialize a basic dataset
    dataset = tf.data.TFRecordDataset([tfrecods_file]).map(
      partial(decode_protobuf,
              cnk=cnk,
              ck2=ck2,
              num_atom_types=num_atom_types,
              atomic_forces=FLAGS.forces,
              num_f_components=num_f_components,
              num_entries=num_entries),
      num_parallel_calls=num_parallel_calls,
    )

    # Repeat the dataset
    dataset = dataset.repeat(count=num_epochs)

    # Shuffle it if needed
    if shuffle:
      min_queue_examples = int(dataset_size * 0.4) + 3 * batch_size
      dataset = dataset.shuffle(buffer_size=min_queue_examples, seed=SEED)

    # Setup the batch
    dataset = dataset.batch(batch_size)

    tf.logging.info("The input pipeline is initialized.")
    tf.logging.info('BATCH_SIZE          = {}'.format(batch_size))
    tf.logging.info('NUM_PARAELLEL_CALLS = {}'.format(num_parallel_calls))
    tf.logging.info('NUM_EPOCHS          = {}'.format(num_epochs))

    iterator = dataset.make_one_shot_iterator()
    return iterator.get_next()



# coding=utf-8
"""
This script is used to infer the neural network.
"""
from __future__ import print_function, absolute_import

from functools import partial

import numpy as np
import tensorflow as tf
from tensorflow.contrib.opt import NadamOptimizer

from . import pipeline
from .constants import SEED
from .constants import VARIABLE_MOVING_AVERAGE_DECAY, LOSS_MOVING_AVERAGE_DECAY
from .summary_utils import add_total_norm_summaries, add_variable_summaries
from .inference import inference_energy, inference_forces
from .utils import lrelu, selu, reduce_l2_norm
from .utils import selu_initializer, msra_initializer
from .amsgrad import AmsGrad

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


FLAGS = tf.app.flags.FLAGS

# Basic model parameters.
tf.app.flags.DEFINE_integer('batch_size', 50,
                            """Number of structures to process in a batch.""")

# Setup the structure of the model
tf.app.flags.DEFINE_string('conv_sizes', '128,64,32',
                           """Comma-separated integers as the sizes of the 
                           convolution layers.""")
tf.app.flags.DEFINE_string('initial_one_body_weights', None,
                           """Comma-separated floats as the initial one-body 
                           weights. Defaults to `ones_initialier`.""")
tf.app.flags.DEFINE_boolean('fixed_one_body', False,
                            """Make the one-body weights fixed.""")
tf.app.flags.DEFINE_integer("trainable_k_max", 3,
                            """Set the trainable k_max.""")
tf.app.flags.DEFINE_string('activation_fn', "tanh",
                           """Set the activation function for conv layers.""")
tf.app.flags.DEFINE_float('alpha', 0.01,
                          """The alpha value of the leaky relu.""")
tf.app.flags.DEFINE_string('normalizer', 'bias',
                           """Set the normalizer: 'bias'(default), 'batch_norm', 
                           'layer_norm' or 'None'. """)
tf.app.flags.DEFINE_string('initializer', 'xvaier',
                           """Set the weights initialization method: msra or 
                           xvaier""")

# Setup the total loss function
tf.app.flags.DEFINE_float('f_loss_weight', 1.0,
                          """The weight of the f-loss in total loss.""")
tf.app.flags.DEFINE_float('y_loss_weight', 1.0,
                          """The weight of the y-loss in total loss.""")
tf.app.flags.DEFINE_boolean('mse', False,
                            """Use MSE loss instead of RMSE loss if True.""")
tf.app.flags.DEFINE_float('l2', None,
                          """Set the lambda of the l2 loss. If None, 
                          l2 regularizer is disabled.""")

# Setup the learning rate. By default the Adam optimizer and constant initial
# learning rate are adopted.
tf.app.flags.DEFINE_float('learning_rate', 0.0001,
                          """The learning rate for minimizing energy.""")
tf.app.flags.DEFINE_string('learning_rate_decay', None,
                           """The decay function. Default is None. Available
                           options are: exponential, inverse_time and 
                           natrual_exp.""")
tf.app.flags.DEFINE_float('learning_rate_decay_factor', 0.95,
                          """A Python number.  The decay rate.""")
tf.app.flags.DEFINE_float('learning_rate_decay_step', 1000,
                          """How often to apply decay.""")
tf.app.flags.DEFINE_boolean('staircase', False,
                            """Boolean.  If `True` decay the learning rate at 
                            discrete intervals""")
tf.app.flags.DEFINE_float('min_learning_rate', None,
                          """Setup the minimum learning rate.""")

# Setup the SGD optimizer.
tf.app.flags.DEFINE_string('optimizer', 'adam',
                           """Set the optimizer to use. Avaible optimizers are:
                           adam, nadam, rmsprop, adadelta""")
tf.app.flags.DEFINE_float('beta1', 0.9,
                          """The beta1 of Adam/Nadam""")
tf.app.flags.DEFINE_float('rho', 0.95,
                          """The rho of Adadelta.""")
tf.app.flags.DEFINE_float('rmsprop_decay', 0.9,
                          """The decay factor of RMSProp.""")
tf.app.flags.DEFINE_float('rmsprop_momentum', 0.0,
                          """The momentum factor for the RMSProp optimizer.""")


def get_learning_rate(global_step):
  """
  Return the tensor of learning rate.

  Args:
    global_step: an `int64` tensor as the global step.

  Returns:
    lr: a `float32` tensor as the learning rate.

  """
  if FLAGS.learning_rate_decay is None:
    learning_rate = tf.constant(
      FLAGS.learning_rate, dtype=tf.float32, name="raw_learning_rate")
  else:
    if FLAGS.learning_rate_decay == 'exponential':
      lr = tf.train.exponential_decay
    elif FLAGS.learning_rate_decay == 'inverse_time':
      lr = tf.train.inverse_time_decay
    elif FLAGS.learning_rate_decay == 'natural_exp':
      lr = tf.train.natural_exp_decay
    else:
      raise ValueError(
        'Supported decay functions: exponential, inverse_time, natural_exp')
    learning_rate = lr(FLAGS.learning_rate, global_step=global_step,
                       decay_rate=FLAGS.learning_rate_decay_factor,
                       decay_steps=FLAGS.learning_rate_decay_step,
                       staircase=FLAGS.staircase, name="raw_lr")

  min_learning_rate = tf.constant(
    FLAGS.min_learning_rate or 0.0, name="minimum_lr")
  learning_rate = tf.maximum(learning_rate, min_learning_rate, "adjusted_lr")
  tf.summary.scalar('lr', learning_rate)
  return learning_rate


def get_optimizer(learning_rate):
  """
  Return the tensor of SGD optimizer.

  Args:
    learning_rate: a `float32` tensor as the learning rate.

  Returns:
    optimizer: an optimizer.

  """

  if FLAGS.optimizer == 'adam':
    return tf.train.AdamOptimizer(
      learning_rate=learning_rate, beta1=FLAGS.beta1)
  elif FLAGS.optimizer == 'nadam':
    return NadamOptimizer(learning_rate=learning_rate, beta1=FLAGS.beta1)
  elif FLAGS.optimizer == 'adadelta':
    return tf.train.AdadeltaOptimizer(
      learning_rate=learning_rate, rho=FLAGS.rho)
  elif FLAGS.optimizer == 'rmsprop':
    return tf.train.RMSPropOptimizer(
      learning_rate=learning_rate,
      decay=FLAGS.rmsprop_decay,
      momentum=FLAGS.rmsprop_momentum)
  elif FLAGS.optimizer == 'amsgrad':
    return AmsGrad(learning_rate=learning_rate, beta1=FLAGS.beta1)
  else:
    raise ValueError(
      "Supported SGD optimizers: adam, nadam, adadelta, rmsprop, amsgrad")


def get_activation_fn(name='lrelu'):
  """
  Return a callable activation function.

  Args:
    name: a `str` as the name of the activation function.

  Returns:
    fn: a `Callable` as the activation function.

  """
  if name.lower() == 'tanh':
    return tf.nn.tanh
  elif name.lower() == 'relu':
    return tf.nn.relu
  elif name.lower() == 'lrelu':
    return lrelu
  elif name.lower() == 'softplus':
    return tf.nn.softplus
  elif name.lower() == 'sigmoid':
    return tf.nn.sigmoid
  elif name.lower() == 'elu':
    return tf.nn.elu
  elif name.lower() == 'selu':
    return selu
  else:
    raise ValueError("The %s activation is not supported!".format(name))


class BatchIndex:
  """
  The indices for manipulating the tuple from `get_batch`.
  """
  inputs = 0
  y_true = 1
  occurs = 2
  weights = 3
  loss_weight = 4
  f_true = 5
  coefficients = 6
  indexing = 7


def kcnn(inputs, occurs, weights, split_dims=(), num_atom_types=None,
         kbody_terms=(), is_training=True, reuse=False, num_kernels=None,
         verbose=True, one_body_weights=None, atomic_forces=False,
         coefficients=None, indexing=None, add_summary=True):
  """
  Inference the model of `KCNN`.

  Args:
    inputs: a Tensor of shape `[-1, 1, -1, D]` as the inputs.
    occurs: a Tensor of shape `[-1, num_atom_types]` as the number of occurances
      of each type of atom.
    weights: a Tensor of shape `[-1, -1, D, -1]` as the weights of the k-body
      contribs.
    split_dims: a `List[int]` or a 1D Tensor of `int` as the dims to split the
      input feature matrix.
    num_atom_types: a `int` as the number of atom types.
    kbody_terms: a `List[str]` as the names of the k-body terms.
    is_training: a `bool` type placeholder indicating whether this inference is
      for training or not.
    reuse: a `bool` indicating whether we should reuse variables or not.
    verbose: boolean indicating whether the layers shall be printed or not.
    num_kernels: a `Tuple[int]` as the number of kernels of the convolution
      layers. This also determines the number of layers in each atomic network.
    one_body_weights: a 1D array of shape `[nat, ]` as the initial
      weights of the one-body kernel.
    atomic_forces: a `bool` indicating whether the atomic forces should be
      trained or not.
    coefficients: a 3D Tensor as the auxiliary coefficients for computing atomic
      forces.
    indexing: a 3D Tensor as the indexing matrix for force compoenents.
    add_summary: a `bool` indicating whether we should add summaries for
      tensors or not.

  Returns:
    y_calc: a tensor of shape `(-1, )` as the predicted total energy.
    f_calc: a tensor of shape `(-1, 3 * num_atoms)` as the predicted forces.
    num_atoms: a tensor of shape `(-1, )` as the total number of atoms for each
      configuration.

  """

  num_kernels = num_kernels or (40, 50, 60, 40)

  activation_fn = get_activation_fn(FLAGS.activation_fn)
  if FLAGS.activation_fn == 'lrelu':
    activation_fn = partial(activation_fn, alpha=FLAGS.alpha)

  weights_initializer = None

  if FLAGS.activation_fn == 'selu':
    weights_initializer = selu_initializer(seed=SEED)
  elif FLAGS.initializer == 'msra':
    weights_initializer = msra_initializer(seed=SEED)

  trainable = not FLAGS.fixed_one_body
  trainable_k_max = FLAGS.trainable_k_max

  if FLAGS.normalizer.lower() == 'none':
    normalizer = None
  else:
    normalizer = FLAGS.normalizer

  if FLAGS.initial_one_body_weights is not None:
    one_body_weights = FLAGS.initial_one_body_weights
    one_body_weights = np.array([float(x) for x in one_body_weights.split(",")])
    if num_atom_types > 1 and len(one_body_weights) == 1:
      one_body_weights = np.ones(
        num_atom_types, dtype=np.float32) * one_body_weights[0]

  with tf.variable_scope("kCON"):
    y_calc, _ = inference_energy(
      inputs,
      occurs,
      weights,
      split_dims,
      num_atom_types=num_atom_types,
      kbody_terms=kbody_terms,
      is_training=is_training,
      reuse=reuse,
      num_kernels=num_kernels,
      activation_fn=activation_fn,
      normalizer=normalizer,
      weights_initializer=weights_initializer,
      one_body_weights=one_body_weights,
      verbose=verbose,
      trainable_one_body=trainable,
      trainable_k_max=trainable_k_max,
      summary=add_summary,
    )

    if atomic_forces:
      f_calc = inference_forces(
        y_total=y_calc,
        inputs=inputs,
        coefficients=coefficients,
        indexing=indexing,
        summary=add_summary
      )
    else:
      f_calc = None

    n_atom = tf.squeeze(
      tf.reduce_sum(occurs, axis=-1, name="sum_of_atoms"), name="num_atoms")
    n_atom.set_shape([None])

  return y_calc, f_calc, n_atom


def extract_configs(configs, for_training=True):
  """
  Extract the config of a dataset.

  Args:
    configs: a `dict` as the configs of a dataset.
    for_training: a `bool` indicating whether the configs should be used for
      training or not.

  Returns:
    params: a `dict` as the parameters for inference.

  """

  # Extract the constant configs from the dict
  split_dims = np.asarray(configs["split_dims"], dtype=np.int64)
  num_atom_types = configs["num_atom_types"]
  kbody_terms = [term.replace(",", "") for term in configs["kbody_terms"]]
  num_kernels = [int(units) for units in FLAGS.conv_sizes.split(",")]
  atomic_forces = configs.get("atomic_forces_enabled", False)

  # The last weight corresponds to the average contribs from k_max-body terms.
  weights = np.array(configs["initial_one_body_weights"], dtype=np.float32)
  if len(weights) == 0:
    weights = np.ones((num_atom_types, ), dtype=np.float32)
  else:
    weights = weights[:num_atom_types]

  # Create the parameter dict and the feed dict
  params = dict(split_dims=split_dims, kbody_terms=kbody_terms,
                is_training=for_training, one_body_weights=weights,
                num_atom_types=num_atom_types, num_kernels=num_kernels,
                atomic_forces=atomic_forces, verbose=True)
  return params


def kcnn_from_dataset(dataset_name, for_training=True, num_epochs=None,
                      **kwargs):
  """
  Inference a kCON model for the given dataset.

  Args:
    dataset_name: a `str` as the name of the dataset.
    for_training: a `bool` indicating whether this inference is for training or
      evaluation.
    num_epochs: an `int` as the maximum number of epochs to run.
    kwargs: additional key-value parameters.

  Returns:
    y_calc: a `float32` Tensor of shape `(-1, )` as the predicted total energy.
    y_true: a `float32` Tensor of shape `(-1, )` as the true energy.
    y_weight: a `float32` Tensor of shape `(-1, )` as the weights for computing
      weighted RMSE loss.
    f_calc: a `float32` Tensor of shape `(-1, -1)` as the predicted atomic
      forces.
    f_true: a `float32` Tensor of shape `(-1, -1)` as the true atomic forces.
    n_atom: a tensor of shape `(-1, )` as the total number of atoms for each
      configuration.

  """
  batch = pipeline.next_batch(
    for_training=for_training,
    dataset_name=dataset_name,
    shuffle=for_training,
    num_epochs=num_epochs,
    batch_size=FLAGS.batch_size,
  )
  configs = pipeline.get_configs(
    for_training=for_training, dataset_name=dataset_name
  )
  params = extract_configs(configs, for_training=for_training)
  for key, val in kwargs.items():
    if key in params:
      params[key] = val
    else:
      tf.logging.warning("Unrecognized key={}".format(key))
  y_true = batch[BatchIndex.y_true]
  y_weight = batch[BatchIndex.loss_weight]
  f_calc = None
  f_true = None

  if not params["atomic_forces"]:
    y_calc, _, n_atom = kcnn(
      batch[BatchIndex.inputs],
      batch[BatchIndex.occurs],
      batch[BatchIndex.weights],
      **params
    )
  else:
    y_calc, f_calc, n_atom = kcnn(
      batch[BatchIndex.inputs],
      batch[BatchIndex.occurs],
      batch[BatchIndex.weights],
      coefficients=batch[BatchIndex.coefficients],
      indexing=batch[BatchIndex.indexing],
      **params,
    )
    f_true = batch[BatchIndex.f_true]

  return y_calc, y_true, y_weight, f_calc, f_true, n_atom


def add_l2_regularizer(total_loss, eta):
  """
  Add l2 regularizer to the total loss.

  Args:
    total_loss: a `float32` tensor as the total loss.
    eta: a `float32` tensor as the strength of the l2. `eta` is used to replace
      `lambda` in the formula because `lambda` is a Python key word.

  Returns:
    total_loss: a `float32` tensor as the regularized total loss.

  """
  with tf.name_scope("L2"):
    for var in tf.trainable_variables():
      if 'bias' in var.op.name:
        continue
      # L2 loss will not include the one-body weights.
      if var.op.name.startswith('one-body'):
        continue
      l2 = tf.nn.l2_loss(var, name=var.op.name + "/l2")
      tf.add_to_collection('l2_loss', l2)
    l2_loss = tf.add_n(tf.get_collection('l2_loss'), name='l2_raw')
    l2_loss = tf.multiply(l2_loss, eta, name='loss')

  tf.summary.scalar('l2_loss', l2_loss)
  return tf.add(total_loss, l2_loss, name='total_loss_and_l2')


def get_y_loss(y_true, y_calc, weights=None):
  """
  Return the total loss tensor of energy only.

  Args:
    y_true: a `float32` tensor of shape `[-1, ]` as the true energies.
    y_calc: a `float32` tensor of shape `[-1, ]` as the computed energies.
    weights: the weights for the energies.

  Returns:
    loss: a `float32` scalar tensor as the total loss.

  """
  return _get_rmse_loss(y_true, y_calc, weights=weights, scope="yRMSE")


def get_f_loss(f_true, f_calc):
  """
  Return the total loss tensor of forces only.

  Args:
    f_true: a `float32` tensor of shape `[-1, -1]` as the true forces.
    f_calc: a `float32` tensor of shape `[-1, -1]` as the computed forces.

  Returns:
    f_loss: a `float32` scalar tensor as the total loss.

  """
  return _get_rmse_loss(f_true, f_calc, scope="fRMSE", summary_norms=True)


def _get_rmse_loss(true, calc, weights=None, scope=None, summary_norms=False):
  """
  Return the total loss tensor.

  Args:
    true: a `float32` tensor as the true values.
    calc: a `float32` tensor as the computed values. This tensor must has the
      same shape with `true`.
    weights: a tensor as the weights for the mean squared errors.
    scope: a `str` as the name scope.
    summary_norms: a `bool` indicating whether we should summary the norms of
      the values or not.

  Returns:
    total_loss: a `float32` scalar tensor as the total loss.

  """

  with tf.name_scope(scope):
    if weights is None:
      weights = tf.constant(1.0, name='weight')
    loss = tf.losses.mean_squared_error(
      true, calc, scope="MSE", loss_collection=None, weights=weights)
    if not FLAGS.mse:
      loss = tf.sqrt(loss, name="RMSE")
    tf.add_to_collection('losses', loss)
    total_loss = tf.add_n(tf.get_collection('losses'), name='total')

    if FLAGS.l2 is not None:
      total_loss = add_l2_regularizer(total_loss, eta=FLAGS.l2)

    if summary_norms:
      l2_true = reduce_l2_norm(true, name="l2_true")
      l2_calc = reduce_l2_norm(calc, name="l2_calc")
      with tf.name_scope("magnitudes"):
        tf.summary.scalar('true', l2_true)
        tf.summary.scalar('calc', l2_calc)

    return total_loss


def get_yf_joint_loss(y_true, y_calc, f_true, f_calc):
  """
  Return the joint total loss tensor.

  Args:
    y_true: a `float32` tensor of shape `[-1, ]` the true energies.
    y_calc: a `float32` tensor of shape `[-1, ]` as the energies.
    f_true: a `float32` tensor of shape `[-1, -1]` as the true forces.
    f_calc: a `float32` tensor of shape `[-1, -1]` as the computed forces.

  Returns:
    loss: a `float32` scalar tensor as the total loss.
    y_loss: a `float32` scalar tensor as the energy loss.
    f_loss: a `float32` scalar tensor as the forces loss.

  """
  with tf.name_scope("yfRMSE"):

    with tf.name_scope("forces"):
      f_loss = tf.losses.mean_squared_error(
        f_true,
        f_calc,
        scope="fMSE",
        loss_collection=None,
      )
      if not FLAGS.mse:
        f_loss = tf.sqrt(f_loss, name="fRMSE")
      f_loss = tf.multiply(f_loss, FLAGS.f_loss_weight, name="f_loss")

      with tf.name_scope("magnitudes"):
        tf.summary.scalar('true', reduce_l2_norm(f_true, name="l2_true"))
        tf.summary.scalar('calc', reduce_l2_norm(f_calc, name="l2_calc"))

    with tf.name_scope("energy"):
      y_loss = tf.losses.mean_squared_error(
        y_true,
        y_calc,
        scope="yMSE",
        loss_collection=None,
      )
      if not FLAGS.mse:
        y_loss = tf.sqrt(y_loss, name="yRMSE")
      y_loss = tf.multiply(y_loss, FLAGS.y_loss_weight, name="y_loss")

    with tf.name_scope("losses"):
      tf.summary.scalar("y", y_loss)
      tf.summary.scalar("f", f_loss)

    loss = tf.add(f_loss, y_loss, name="together")
    tf.add_to_collection("losses", loss)
    total_loss = tf.add_n(tf.get_collection("losses"), name="total")

    if FLAGS.l2 is not None:
      total_loss = add_l2_regularizer(total_loss, eta=FLAGS.l2)

    return total_loss, y_loss, f_loss


def get_amp_yf_joint_loss(y_true, y_calc, f_true, f_calc, n_atom):
  """
  The Amp joint loss function.

  Args:
    y_true: a `float32` tensor of shape `(-1, )` the true energies.
    y_calc: a `float32` tensor of shape `(-1, )` as the energies.
    f_true: a `float32` tensor of shape `(-1, -1)` as the true forces.
    f_calc: a `float32` tensor of shape `(-1, -1)` as the computed forces.
    n_atom: a `int32` tensor of shape `(-1, )` as the number of atoms for each
      configuration.

  Returns:
    loss: a `float32` scalar tensor as the total loss.
    y_loss: a `float32` scalar tensor as the energy loss.
    f_loss: a `float32` scalar tensor as the forces loss.

  """

  with tf.name_scope("yf"):

    with tf.name_scope("energy"):
      y_true = tf.divide(y_true, n_atom, name="true/per_atom")
      y_calc = tf.divide(y_calc, n_atom, name="calc/per_atom")
      y_diff = tf.squared_difference(y_true, y_calc, name="squared")
      y_loss = tf.reduce_sum(y_diff, name='loss')

    with tf.name_scope("forces"):
      alpha = tf.constant(1.0, dtype=tf.float32, name="alpha")
      f_diff = tf.squared_difference(f_true, f_calc, name="squared")
      f_loss = tf.reduce_sum(tf.reduce_mean(f_diff, axis=1))
      f_loss = tf.multiply(f_loss, alpha, name="loss")

      with tf.name_scope("magnitudes"):
        tf.summary.scalar('true', reduce_l2_norm(f_true, name="l2_true"))
        tf.summary.scalar('calc', reduce_l2_norm(f_calc, name="l2_calc"))

    with tf.name_scope("losses"):
      tf.summary.scalar('y', y_loss)
      tf.summary.scalar('f', f_loss)

    loss = tf.multiply(0.5, tf.add(y_loss, f_loss, name="add"), name="loss")
    tf.add_to_collection('losses', loss)
    total_loss = tf.add_n(tf.get_collection('losses'), name="total")

    if FLAGS.l2 is not None:
      total_loss = add_l2_regularizer(total_loss, eta=FLAGS.l2)

    return total_loss, y_loss, f_loss


def _add_loss_summaries(total_loss):
  """Add summaries for losses in KCNN model.

  Generates moving average for all losses and associated summaries for
  visualizing the performance of the network.

  Args:
    total_loss: Total loss from loss().
  
  Returns:
    loss_averages_op: op for generating moving averages of losses.
  
  """

  # Compute the moving average of all individual losses and the total loss.
  loss_averages = tf.train.ExponentialMovingAverage(LOSS_MOVING_AVERAGE_DECAY,
                                                    name='avg')
  losses = tf.get_collection('losses')
  loss_averages_op = loss_averages.apply(losses + [total_loss])

  # Attach a scalar summary to all individual losses and the total loss; do the
  # same for the averaged version of the losses.
  for l in losses + [total_loss]:
    # Name each loss as '(raw)' and name the moving average version of the loss
    # as the original loss name.
    tf.summary.scalar(l.op.name + '_raw', l)
    tf.summary.scalar(l.op.name, loss_averages.average(l))

  return loss_averages_op


def get_joint_loss_train_op(total_loss, global_step):
  """
  Train the model by minimizing the joint total loss.

  Create an optimizer and apply to all trainable variables. Add moving
  average for all trainable variables.

  Args:
    total_loss: a `float32` Tensor as the joint total loss.
    global_step: Integer Variable counting the number of training steps
      processed.

  Returns:
    train_op: the op for training.

  """

  # Generate moving averages of all losses and associated summaries.
  loss_averages_op = _add_loss_summaries(total_loss)

  # Add the update ops if batch_norm is True.
  # If we don't include the update ops as dependencies on the train step, the
  # batch_normalization layers won't update their population statistics, which
  # will cause the model to fail at inference time.
  dependencies = [loss_averages_op]
  if FLAGS.normalizer and FLAGS.normalizer == 'batch_norm':
    dependencies.extend(tf.get_collection(tf.GraphKeys.UPDATE_OPS))

  # Compute gradients.
  with tf.name_scope("Optimizer"):
    with tf.control_dependencies(dependencies):
      lr = get_learning_rate(global_step=global_step)
      opt = get_optimizer(learning_rate=lr)
      grads = opt.compute_gradients(total_loss)

  # Add histograms for grandients
  add_total_norm_summaries(grads, "joint", only_summary_total=False)

  # Apply gradients.
  apply_gradient_op = opt.apply_gradients(
    grads,
    global_step=global_step,
    name="apply_grads"
  )

  # Add histograms for trainable variables.
  add_variable_summaries()

  # Track the moving averages of all trainable variables.
  with tf.name_scope("average"):
    variable_averages = tf.train.ExponentialMovingAverage(
      VARIABLE_MOVING_AVERAGE_DECAY, global_step)
    variables_averages_op = variable_averages.apply(tf.trainable_variables())

  with tf.control_dependencies([apply_gradient_op, variables_averages_op]):
    train_op = tf.no_op(name='train')

  return train_op


def sum_of_gradients(grads_of_losses):
  """
  Calculate the total gradient from `grad_and_vars` of different losses.

  Args:
    grads_of_losses: a list of lists of (gradient, variable) tuples.

  Returns:
    List of pairs of (gradient, variable) as the total gradient.

  """
  # Merge gradients
  sum_grads = []
  for grad_and_vars in zip(*grads_of_losses):
    # Note that each grad_and_vars looks like the following:
    #   ((grad0_gpu0, var0_gpu0), ... , (grad0_gpuN, var0_gpuN))
    grads = []
    for g, _ in grad_and_vars:
      if g is None:
        continue

      # Add 0 dimension to the gradients to represent the tower.
      expanded_g = tf.expand_dims(g, 0)

      # Append on a 'tower' dimension which we will average over below.
      grads.append(expanded_g)

    v = grad_and_vars[0][1]

    # If the grads are all None, we just return a None grad.
    if len(grads) == 0:
      grad_and_var = (None, v)

    else:
      # Average over the 'tower' dimension.
      grad = tf.concat(grads, 0)
      grad = tf.reduce_sum(grad, 0)

      # Keep in mind that the Variables are redundant because they are shared
      # across towers. So .. we will just return the first tower's pointer to
      # the Variable.
      grad_and_var = (grad, v)
    sum_grads.append(grad_and_var)

  return sum_grads


def get_yf_train_op(total_loss, y_loss, f_loss, global_step):
  """
  An alternative implementation of building the `train_op`.

  Args:
    total_loss: a `float32` scalar tensor as the total loss. `total_loss` should
      be the sum of `y_loss` and `f_loss`. In this function `total_loss` is used
      for summary only.
    y_loss: a `float32` scalar tensor as the energy loss.
    f_loss: a `float32` scalar tensor as the forces loss.
    global_step: Integer Variable counting the number of training steps
      processed.

  Returns:
    train_op: the op for training.

  """
  total_loss_op = _add_loss_summaries(total_loss)

  # Add the update ops if batch_norm is True.
  # If we don't include the update ops as dependencies on the train step, the
  # batch_normalization layers won't update their population statistics, which
  # will cause the model to fail at inference time.
  dependencies = [total_loss_op]
  if FLAGS.normalizer and FLAGS.normalizer == 'batch_norm':
    dependencies.extend(tf.get_collection(tf.GraphKeys.UPDATE_OPS))

  with tf.name_scope("Optimizer"):
    with tf.control_dependencies(dependencies):
      lr = get_learning_rate(global_step=global_step)
      opt = get_optimizer(learning_rate=lr)
      assert isinstance(opt, tf.train.Optimizer)

  with tf.name_scope("Gradients"):
    with tf.name_scope("y"):
      dydw = opt.compute_gradients(y_loss)
      add_total_norm_summaries(dydw, "y")

    with tf.name_scope("f"):
      dfdw = opt.compute_gradients(f_loss)
      add_total_norm_summaries(dfdw, "f")

  # Merge the gradients
  sum_grads = sum_of_gradients((dydw, dfdw))
  add_total_norm_summaries(sum_grads, "yf", only_summary_total=False)

  # Apply the gradients to variables
  apply_gradient_op = opt.apply_gradients(
    sum_grads, global_step=global_step, name="apply_merged"
  )

  # Add histograms for trainable variables.
  add_variable_summaries()

  # Track the moving averages of all trainable variables.
  with tf.name_scope("average"):
    variable_averages = tf.train.ExponentialMovingAverage(
      VARIABLE_MOVING_AVERAGE_DECAY, global_step)
    variables_averages_op = variable_averages.apply(tf.trainable_variables())

  with tf.control_dependencies([apply_gradient_op, variables_averages_op]):
    train_op = tf.no_op(name='train')

  return train_op


def test_sum_of_gradients():
  """
  Test the function `sum_of_gradients`.
  """
  with tf.Graph().as_default():
    a = tf.get_variable('a', shape=(), dtype=tf.float32,
                        initializer=tf.zeros_initializer())
    b = tf.get_variable('b', shape=(), dtype=tf.float32,
                        initializer=tf.zeros_initializer())
    x = tf.placeholder(tf.float32, shape=(None, ), name='x')
    y = tf.placeholder(tf.float32, shape=(None, ), name='y')

    f_calc = a * x + b * tf.pow(y, 2)
    f_true = tf.placeholder(tf.float32, shape=(None, ), name='f_true')
    f_loss = tf.sqrt(tf.losses.mean_squared_error(f_true, f_calc))
    g_calc = b * y
    g_true = tf.placeholder(tf.float32, shape=(None, ), name='g_true')
    g_loss = tf.sqrt(tf.losses.mean_squared_error(g_true, g_calc))
    h_loss = f_loss + g_loss

    data = [[2.0, 3.0, 31.0, 9.0],
            [1.0, 1.0, 5.0, 3.0]]
    data = np.array(data, dtype=np.float32)

    opt = tf.train.GradientDescentOptimizer(learning_rate=0.1)

    with tf.name_scope("h"):
      raw_grads = opt.compute_gradients(h_loss)

    with tf.name_scope("separated"):
      with tf.name_scope("f"):
        dfdx = opt.compute_gradients(f_loss)
      with tf.name_scope("g"):
        dgdx = opt.compute_gradients(g_loss)
      sum_grads = sum_of_gradients((dfdx, dgdx))

    with tf.Session() as sess:
      tf.global_variables_initializer().run()
      feed_dict = {x: data[:, 0],
                   y: data[:, 1],
                   f_true: data[:, 2],
                   g_true: data[:, 3]}
      grad1 = sess.run(raw_grads, feed_dict=feed_dict)
      grad2 = sess.run(sum_grads, feed_dict=feed_dict)
      assert abs(grad1[0][0] - grad2[0][0]) < 0.001
      assert abs(grad1[1][0] - grad2[1][0]) < 0.001



import os
import time
import logging
from abc import ABCMeta


class kconMD_CS(metaclass=ABCMeta):
    def __init__(self, server_path="/tmp/kconmd.server",
                 client_path="/tmp/kconmd.client", sleeptime=1):
        self.server_path = server_path
        self.client_path = client_path
        self.sleeptime = sleeptime
        self.rf = None
        self.wf = None

    def _message(self, message):
        return f"{self.CStype.upper()}: {message}"

    def mkfifos(self):
        for path in (self.server_path, self.client_path):
            if os.path.exists(path):
                os.remove(path)
            os.mkfifo(path)

    def openfifo(self, path, readonly=True):
        if readonly:
            f = os.open(path, os.O_RDONLY)
        else:
            f = os.open(path, os.O_SYNC | os.O_CREAT | os.O_RDWR)
        return f

    def sendmessage(self, message):
        if self.wf is None:
            self.wf = self.openfifo(self.write_path, readonly=False)
        os.write(self.wf, bytes(message, 'UTF-8'))
        logging.info(self._message(f"Send message: {message}"))

    def response(self):
        while True:
            if self.rf is None:
                self.rf = self.openfifo(self.read_path, readonly=True)
            s = str(os.read(self.rf, 1024), 'UTF-8')
            if len(s) == 0:
                time.sleep(self.sleeptime)
                continue
            logging.info(self._message(f"Receive message: {s}"))
            if "Exit" in s:
                logging.info(self._message("EXIT"))
                break
            self.handleMessage(s)

    def handleMessage(self, message):
        pass


class kconMD_server(kconMD_CS):
    def __init__(self, kconMD, server_path="/tmp/kconmd.server",
                 client_path="/tmp/kconmd.client", sleeptime=1):
        kconMD_CS.__init__(self, server_path=server_path,
                           client_path=client_path, sleeptime=sleeptime)
        self.mkfifos()
        self.kconMD = kconMD
        self.kconMD.initcf()
        logging.info(self._message("Initialization is complete."))

    @property
    def printforce(self):
        return self.kconMD.printforce

    @property
    def read_path(self):
        return self.server_path

    @property
    def write_path(self):
        return self.client_path

    @property
    def CStype(self):
        return 'server'

    def handleMessage(self, message):
        if "Printforce" in message:
            logging.info(self._message("Print Force"))
            self.printforce()
            self.sendmessage("Exit")


class kconMD_client(kconMD_CS):
    def __init__(self, server_path="/tmp/kconmd.server",
                 client_path="/tmp/kconmd.client", sleeptime=1):
        kconMD_CS.__init__(self, server_path=server_path,
                           client_path=client_path, sleeptime=sleeptime)

    @property
    def read_path(self):
        return self.client_path

    @property
    def write_path(self):
        return self.server_path

    @property
    def CStype(self):
        return 'client'

    def printforce(self):
        self.sendmessage("Printforce")
        self.response()

    def exitserver(self):
        self.sendmessage("Exit")



#!/usr/bin/env python3
# Jinzhe Zeng
# 2018/08/21
from multiprocessing import Pool, cpu_count

import numpy as np
from ase.io import read

from ..kcnn.predictor import KcnnPredictor
from .f_vdw import f_vdw
from .feed import Feed


class ComputeForces:
    def __init__(
            self, pbfilename, cell=None,
            pbc=False, cutoff=6, vdw=False, nproc=None):
        self.clf = KcnnPredictor(pbfilename, fixed=True)
        self.cell = cell if cell else [0, 0, 0]
        self.pbc = pbc
        self.cutoff = cutoff
        self.maxatoms = self.clf.transformer.max_occurs
        self.feedobject = Feed(self.clf.transformer, self.cutoff)
        self.nproc = nproc if nproc else cpu_count()
        self._pool = Pool(self.nproc)
        self.vdw = vdw

    def predictforces(self, atoms):
        self.N = len(atoms)
        forces = np.zeros((self.N, 3))
        # NN forces
        feeds = self.feedobject.runmp(atoms, self._pool)
        for i, feed in enumerate(feeds):
            forces[i] = np.array(self.computeforcefromfeed(
                feed)[0][3*feed['index']:3*feed['index']+3])
        # vdw forces
        if self.vdw:
            forces += f_vdw(cutoff=self.cutoff).calculate_forces(atoms)
        # Make the resultant force equal to 0
        if np.abs(np.sum(forces)) > 0:
            forces -= np.abs(forces)/np.sum(np.abs(forces),
                                            0)*np.sum(forces, 0)
        return forces

    def computeforcefromfeed(self, feed):
        feed_dict = {self.clf._placeholder_inputs: feed["inputs"],
                     self.clf._placeholder_occurs: feed["occurs"],
                     self.clf._placeholder_weights: feed["weights"],
                     self.clf._placeholder_split_dims: feed["split_dims"],
                     self.clf._placeholder_coefficients: feed["coefficients"],
                     self.clf._placeholder_indexing: feed["indexing"]}
        f_atomics = self.clf._sess.run(
            self.clf._operator_f_nn, feed_dict=feed_dict)
        return f_atomics

    def predictforcesfromxyz(self, xyzfilename):
        atoms = read(xyzfilename)
        atoms.set_cell(self.cell)
        atoms.set_pbc(self.pbc)
        return self.predictforces(atoms)



# coding=utf-8
"""
Some utility functions.
"""

from __future__ import print_function, absolute_import

import tensorflow as tf
import numpy as np
import logging
import json
import time
from scipy.misc import factorial, comb
from logging.config import dictConfig
from tensorflow.python.framework import ops
from tensorflow.python.ops import math_ops
from tensorflow.contrib.layers import variance_scaling_initializer
from os import getpid
from os.path import join, dirname
from sys import platform
from subprocess import Popen, PIPE
from sys import version_info

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


class Gauss:
  """
  A simple implementation of the Gaussian function.
  """

  def __init__(self, mu, sigma):
    """
    Initialization method.

    Args:
      mu: a `float` as the center of the gaussian.
      sigma: a `float` as the sigma.

    """
    self.mu = mu
    self.sigma = sigma
    self.beta = 1.0 / (2.0 * sigma**2)
    self.c = 1.0 / (sigma * np.sqrt(2.0 * np.pi))

  def __call__(self, x):
    """
    Compute the values at `x`.
    """
    return self.c * np.exp(-self.beta * (x - self.mu)**2)


def compute_n_from_cnk(cnk, k):
  """
  Return the correponding N given C(N, k) and k.

  Args:
    cnk: an `int` as the value of C(N, k).
    k: an `int` as the value of k.

  Returns:
    n: an `int` as the value of N.

  """
  if k == 2:
    return int((1 + np.sqrt(1 + 8 * cnk)) * 0.5)
  else:
    istart = int(np.floor(np.power(factorial(k) * cnk, 1.0 / k)))
    for v in range(istart, istart + k):
      if comb(v, k) == cnk:
        return v
    else:
      raise ValueError(
        "The N for C(N, {}) = {} cannot be solved!".format(k, cnk))


def safe_divide(a, b):
  """
  Safe division while ignoring / 0, div0( [-1, 0, 1], 0 ) -> [0, 0, 0].

  References:
     https://stackoverflow.com/questions/26248654/numpy-return-0-with-divide-by-zero

  """
  with np.errstate(divide='ignore', invalid='ignore'):
    c = np.true_divide(a, b)
    c[~ np.isfinite(c)] = 0  # -inf inf NaN
  return c


def get_atoms_from_kbody_term(kbody_term):
  """
  Return the atoms in the given k-body term.

  Args:
    kbody_term: a `str` as the k-body term.

  Returns:
    atoms: a `list` of `str` as the chemical symbols of the atoms.

  """
  sel = [0]
  for i in range(len(kbody_term)):
    if kbody_term[i].isupper():
      sel.append(i + 1)
    else:
      sel[-1] += 1
  atoms = []
  for i in range(len(sel) - 1):
    atoms.append(kbody_term[sel[i]: sel[i + 1]])
  return atoms


def get_k_from_var(var):
  """
  Get the associated `k` for the given variable.

  Args:
    var: a `tf.Variable`.

  Returns:
    k: an `int`.

  """
  name = var.op.name
  elements = name.split("/")
  if len(elements) >= 4 and elements[0] == 'kCON':
    symbols = get_atoms_from_kbody_term(elements[1])
    return len(symbols) - symbols.count("X")
  else:
    return -1


def lrelu(x, alpha=0.01, name=None):
  """
  The leaky relu activation function.

  `f(x) = alpha * x for x < 0`,
  `f(x) = x for x >= 0`.

  Args:
    x: a `Tensor` with type `float`, `double`, `int32`, `int64`, `uint8`,
      `int16`, or `int8`.
    alpha: a `float32` tensor as the alpha.
    name: a `str` as the name of this op.

  Returns:
    y: a `Tensor` with the same type as `x`.

  """
  with ops.name_scope(name, "LRelu", [x]) as name:
    alpha = ops.convert_to_tensor(alpha, dtype=tf.float32, name="alpha")
    z = math_ops.multiply(alpha, x, "z")
    return math_ops.maximum(z, x, name=name)


def selu(x, name=None):
  """
  The Scaled Exponential Linear Units.

  Args:
    x: a `Tensor` with type `float`, `double`, `int32`, `int64`, `uint8`,
      `int16`, or `int8`.
    name: a `str` as the name of this op.

  Returns:
    y: a `Tensor` with the same type as `x`.

  References:
    https://arxiv.org/pdf/1706.02515.pdf

  """
  with ops.name_scope(name, "SeLU", [x]):
    alpha = 1.6732632423543772848170429916717
    scale = 1.0507009873554804934193349852946
    elu = tf.nn.elu(x, name='elu')
    z = tf.multiply(elu, alpha, name='z')
    return scale * tf.where(x >= 0.0, x, z, name='selu')


def selu_initializer(dtype=tf.float32, seed=None):
  """
  The weights initializer for selu activations.

  Args:
    seed: A Python integer. Used to create random seeds. See
          @{tf.set_random_seed} for behavior.
    dtype: The data type. Only floating point types are supported.

  Returns:
    An initializer for a weight matrix.

  """
  mode = 'FAN_IN'
  return variance_scaling_initializer(
    factor=1.0, mode=mode, dtype=dtype, seed=seed)


def reduce_l2_norm(tensor, name=None):
  """
  Return the mean of the L2 norms along axis 1 of the given tensor.

  Args:
    tensor: a `float32` tensor of rank 2.
    name: a `str` as the name of the op.

  Returns:
    norm: a `float32` tensor as the mean of the L2 norms.

  """
  with ops.name_scope(name, "reduce_norm", [tensor]):
    norms = tf.norm(tensor, axis=1, keep_dims=False, name="norm")
    return tf.reduce_mean(norms, name="mean")


def msra_initializer(dtype=tf.float32, seed=None):
  """
  [Delving Deep into Rectifiers](http://arxiv.org/pdf/1502.01852v1.pdf)

  Args:
    seed: A Python integer. Used to create random seeds. See
          @{tf.set_random_seed} for behavior.
    dtype: The data type. Only floating point types are supported.

  Returns:
    An initializer for a weight matrix.

  """
  return variance_scaling_initializer(seed=seed, dtype=dtype)


def get_xargs(pid=None):
  """
  Return the build and execute command lines from standard input for a process. 
  This is a wrapper of the linux command `xargs -0 <`. maxOS and Win 
  are not supported.
  
  Args:
    pid: an `int` as the target process. 
  
  Returns:
    exe_args: a `str` as the execute command line for process `pid` or None.

  """
  if platform != "linux":
    return None
  if not pid:
    pid = getpid()

  p = Popen("xargs -0 < /proc/{}/cmdline".format(pid), stdout=PIPE, stderr=PIPE,
            shell=True)
  stdout, stderr = p.communicate()
  if len(stderr) > 0:
    return None
  elif version_info > (3, 0):
    return bytes(stdout).decode()
  else:
    return stdout


def set_logging_configs(debug=False, logfile="logfile", is_eval=False):
  """
  Config the logging module.
  """
  if is_eval:
    level = logging.INFO
    handlers = ['normal', 'eval']
  elif debug:
    level = logging.DEBUG
    handlers = ['console', 'normal']
  else:
    level = logging.INFO
    handlers = ['normal']

  LOGGING_CONFIG = {
    "version": 1,
    "formatters": {
      # For normal logs
      'detailed': {
        'format': '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
      },
      # For the console
      'console': {
        'format': '[%(levelname)s] %(message)s'
      },
      # For the evaluations results
      'simple': {
        'format': "%(message)s"
      }
    },
    "handlers": {
      'console': {
        'class': 'logging.StreamHandler',
        'level': logging.DEBUG,
        'formatter': 'console',
      },
      # Redirect all logs to the file `logfile`.
      'normal': {
        'class': 'logging.FileHandler',
        'level': logging.INFO,
        'formatter': 'detailed',
        'filename': logfile,
        'mode': 'a',
      },
      # Redirect the simplfied evaluation results to the file SUMMARY.
      'eval': {
        'class': 'logging.FileHandler',
        'level': logging.CRITICAL,
        'formatter': 'simple',
        'filename': join(dirname(logfile), 'SUMMARY'),
        'mode': 'a',
      }
    },
    "root": {
      'handlers': handlers,
      'level': level,
    },
    "disable_existing_loggers": False
  }
  dictConfig(LOGGING_CONFIG)


def save_training_flags(save_dir, args):
  """
  Save the training flags to the train_dir.
  """
  args["run_flags"] = " ".join(
    ["--{}={}".format(k, v) for k, v in args.items()]
  )
  cmdline = get_xargs()
  if cmdline:
    args["cmdline"] = cmdline
  timestamp = "{}".format(int(time.time()))
  with open(join(save_dir, "flags.{}.json".format(timestamp)), "w+") as f:
    json.dump(args, f, indent=2)



"""Init."""

__author__ = 'Jinzhe Zeng'

import logging

import coloredlogs
from pkg_resources import DistributionNotFound, get_distribution

from .kconmd import kconMD

__all__ = ['kconMD']

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = ''

coloredlogs.install(
    fmt=f'%(asctime)s - kconMD {__version__} - %(levelname)s: %(message)s',
    level=logging.INFO, milliseconds=True)



#!coding=utf-8
"""
The global constants.
"""
from __future__ import print_function, absolute_import

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'

# Set the random state
SEED = 218

# Hartree to kcal/mol
hartree_to_kcal = 627.509

# Hartree to eV
hartree_to_ev = 27.211

# a.u to angstroms
au_to_angstrom = 0.52917721092

# The pyykko radius of each element.
pyykko = {
  'Ac': 1.86, 'Ag': 1.28, 'Al': 1.26, 'Am': 1.66, 'Ar': 0.96, 'As': 1.21,
  'At': 1.47, 'Au': 1.24, 'B': 0.85, 'Ba': 1.96, 'Be': 1.02, 'Bh': 1.41,
  'Bi': 1.51, 'Bk': 1.68, 'Br': 1.14, 'C': 0.75, 'Ca': 1.71, 'Cd': 1.36,
  'Ce': 1.63, 'Cf': 1.68, 'Cl': 0.99, 'Cm': 1.66, 'Co': 1.11, 'Cr': 1.22,
  'Cs': 2.32, 'Cu': 1.12, 'Db': 1.49, 'Ds': 1.28, 'Dy': 1.67, 'Er': 1.65,
  'Es': 1.65, 'Eu': 1.68, 'F': 0.64, 'Fe': 1.16, 'Fm': 1.67, 'Fr': 2.23,
  'Ga': 1.24, 'Gd': 1.69, 'Ge': 1.21, 'H': 0.32, 'He': 0.46, 'Hf': 1.52,
  'Hg': 1.33, 'Ho': 1.66, 'Hs': 1.34, 'I': 1.33, 'In': 1.42, 'Ir': 1.22,
  'K': 1.96, 'Kr': 1.17, 'La': 1.8, 'Li': 1.33, 'Lu': 1.62, 'Md': 1.73,
  'Mg': 1.39, 'Mn': 1.19, 'Mo': 1.38, 'Mt': 1.29, 'N': 0.71, 'Na': 1.55,
  'Nb': 1.47, 'Nd': 1.74, 'Ne': 0.67, 'Ni': 1.1, 'No': 1.76, 'Np': 1.71,
  'O': 0.63, 'Os': 1.29, 'P': 1.11, 'Pa': 1.69, 'Pb': 1.44, 'Pd': 1.2,
  'Pm': 1.73, 'Po': 1.45, 'Pr': 1.76, 'Pt': 1.23, 'Pu': 1.72, 'Ra': 2.01,
  'Rb': 2.1, 'Re': 1.31, 'Rf': 1.57, 'Rh': 1.25, 'Rn': 1.42, 'Ru': 1.25,
  'S': 1.03, 'Sb': 1.4, 'Sc': 1.48, 'Se': 1.16, 'Sg': 1.43, 'Si': 1.16,
  'Sm': 1.72, 'Sn': 1.4, 'Sr': 1.85, 'Ta': 1.46, 'Tb': 1.68, 'Tc': 1.28,
  'Te': 1.36, 'Th': 1.75, 'Ti': 1.36, 'Tl': 1.44, 'Tm': 1.64, 'U': 1.7,
  'V': 1.34, 'W': 1.37, 'Xe': 1.31, 'Y': 1.63, 'Yb': 1.7,
  'Zn': 1.18, 'Zr': 1.54,
  # X represents virtual atoms
  'X': 0.32,
}

# The half length of the Rm of the Lennard-Jones potential.
LJR = 0.5 * 1.122462048309373

# the ghost atom
GHOST = 'X'

# Constants describing the training process.

# The decay to use for the moving average of variable weights.
VARIABLE_MOVING_AVERAGE_DECAY = 0.9999

# The decay to use for the moving average of losses.
LOSS_MOVING_AVERAGE_DECAY = 0.9

# If a model is trained with multiple GPUs, prefix all Op names with tower_name
# to differentiate the operations. Note that this prefix is removed from the
# names of the summaries when visualizing a model.
TOWER_NAME = 'tower'


class KcnnGraphKeys:
  """
  Custom graph collections for kCON.

  * `ENERGY_VARIABLES`: variables that should be updated during energy training.
  * `FORCES_VARIABLES`: variables that should be updated during forces training.

  """
  ENERGY_VARIABLES = 'energy'
  FORCES_VARIABLES = 'forces'



from math import exp

import numpy as np
from ase.io import read


class f_vdw:
    def __init__(self, cutoff=6):
        self.cutoff = cutoff
        self.kcal_to_eV = 1./23.06035

        swa = 0.00
        swb = 10.00
        d1 = swb - swa
        d7 = pow(d1, 7.0)
        swa2 = swa ** 2
        swa3 = swa ** 3
        swb2 = swb ** 2
        swb3 = swb ** 3

        Tap = [0 for i in range(8)]
        Tap[7] = 20.0 / d7
        Tap[6] = -70.0 * (swa + swb) / d7
        Tap[5] = 84.0 * (swa2 + 3.0*swa*swb + swb2) / d7
        Tap[4] = -35.0 * (swa3 + 9.0*swa2*swb + 9.0*swa*swb2 + swb3) / d7
        Tap[3] = 140.0 * (swa3*swb + 3.0*swa2*swb2 + swa*swb3) / d7
        Tap[2] = -210.0 * (swa3*swb2 + swa2*swb3) / d7
        Tap[1] = 140.0 * swa3 * swb3 / d7
        Tap[0] = (-35.0*swa3*swb2*swb2 + 21.0*swa2*swb3 *
                  swb2 + 7.0*swa*swb3*swb3 + swb3*swb3*swb) / d7
        self.Tap = Tap

        # r_vdW D alpha
        self._twbps = {"C": [1.9133, 0.1853, 9.7602],
                       "H": [1.5904, 0.0419, 9.3557],
                       "O": [1.9236, 0.0904, 10.2127],
                       "CH": [1.4000, 0.1219, 9.8442],
                       "HO": [1.6800, 0.0344, 10.3247],
                       "CO": [1.8523, 0.1131, 9.8442]
                       }
        self.twbpname = {
            "C": {"C": "C", "H": "CH", "O": "CO"},
            "H": {"C": "CH", "H": "H", "O": "HO"},
            "O": {"C": "CO", "H": "HO", "O": "O"}}

    def calculate_forces(self, atoms):
        Tap = self.Tap
        n = len(atoms)
        distances = atoms.get_all_distances(mic=True)
        forces = np.zeros((n, n))
        for i in range(n):
            for j in range(i+1, n):
                r_ij = distances[i][j]
                if r_ij <= self.cutoff:
                    r_vdW, D, alpha = self._twbps[self.twbpname[atoms[i].symbol]
                                                  [atoms[j].symbol]]

                    Tap1 = Tap[7] * r_ij + Tap[6]
                    Tap1 = Tap1 * r_ij + Tap[5]
                    Tap1 = Tap1 * r_ij + Tap[4]
                    Tap1 = Tap1 * r_ij + Tap[3]
                    Tap1 = Tap1 * r_ij + Tap[2]
                    Tap1 = Tap1 * r_ij + Tap[1]
                    Tap1 = Tap1 * r_ij + Tap[0]

                    dTap = 7*Tap[7] * r_ij + 6*Tap[6]
                    dTap = dTap * r_ij + 5*Tap[5]
                    dTap = dTap * r_ij + 4*Tap[4]
                    dTap = dTap * r_ij + 3*Tap[3]
                    dTap = dTap * r_ij + 2*Tap[2]
                    dTap += Tap[1]/r_ij

                    exp1 = exp(alpha * (1.0 - r_ij / r_vdW))
                    exp2 = exp(0.5 * alpha * (1.0 - r_ij / r_vdW))

                    e_vdW = D * (exp1 - 2.0 * exp2)

                    CEvd = dTap*e_vdW - Tap1 * D * \
                        (alpha / r_vdW) * (exp1 - exp2) / r_ij

                    forces[i][j] = -CEvd
                    forces[j][i] = -CEvd

        atomforce = np.zeros((n, 3))
        for i in range(n):
            for j in range(i+1, n):
                atomforce_ij = atoms.get_distance(
                    i, j, mic=True, vector=True)/np.linalg.norm(distances[i][j])*forces[i][j]
                atomforce[i] += -atomforce_ij
                atomforce[j] += atomforce_ij

        atomforce *= self.kcal_to_eV
        return atomforce



#!coding=utf-8
"""
This module is used to inference the model of `KCNN`.
"""
from __future__ import print_function, absolute_import

import numpy as np
import tensorflow as tf
from tensorflow.contrib.framework import arg_scope
from tensorflow.contrib.layers import batch_norm, layer_norm
from tensorflow.contrib.layers import conv2d, flatten
from tensorflow.contrib.layers.python.layers import initializers
from tensorflow.python.ops import init_ops
from .constants import KcnnGraphKeys, SEED, GHOST
from .utils import lrelu, get_atoms_from_kbody_term

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


# The decay for batch normalization
BATCH_NORM_DECAY_FACTOR = 0.999


def _inference_kbody_cnn(inputs, kbody_term, is_training, verbose=True,
                         reuse=False, normalizer='bias', activation_fn=lrelu,
                         weights_initializer=None, num_kernels=None,
                         trainable=True):
  """
  Infer the k-body term of `KCNN`.

  Args:
    inputs: a `[-1, 1, -1, C(k, 2)]` Tensor as the inputs for this interaction.
    kbody_term: a `str` as the name of this k-body atomic interaction.
    is_training: a `bool` type placeholder tensor indicating whether this
      inference is for training or not.
    reuse: a `bool` indicating whether we should reuse the variables or not.
    normalizer: a `str` as the normalizer to use. Supported normalizers include:
      'bias' (default), 'batch_norm', 'layer_norm' and None.
    activation_fn: a `Callable` as the activation function for each conv layer.
    weights_initializer: a `Callable` as the function to intialize weights.
    num_kernels: a `List[int]` as the number of kernels.
    trainable: a `bool` indicating whether we should train the variables or not.
    verbose: a `bool`. If Ture, the shapes of the layers will be printed.

  Returns:
    contribs: a Tensor of shape `[-1, 1, N, 1]` as the energy contribs of all
      possible atom combinations.

  """

  if verbose:
    tf.logging.info("Infer the %s term of `KCNN` ..." % kbody_term)

  num_kernels = num_kernels or (40, 50, 60, 40)
  kernel_size = 1
  dtype = tf.float32

  # Setup the initializers and normalization function.
  weights_initializer = weights_initializer or initializers.xavier_initializer(
    seed=SEED, dtype=dtype)
  biases_initializer = None
  normalizer_fn = None
  normalizer_params = {}
  collections = [KcnnGraphKeys.FORCES_VARIABLES,
                 KcnnGraphKeys.ENERGY_VARIABLES]

  if normalizer == 'bias':
    biases_initializer = init_ops.zeros_initializer()
  elif normalizer == 'batch_norm':
    normalizer_fn = batch_norm
    normalizer_params = {"decay": BATCH_NORM_DECAY_FACTOR,
                         "scale": True,
                         "center": True,
                         "scope": "bn",
                         "reuse": reuse,
                         "is_training": is_training,
                         "variables_collections": collections}
  elif normalizer == 'layer_norm':
    biases_initializer = init_ops.zeros_initializer()
    normalizer_fn = layer_norm
    normalizer_params = {"scope": "ln",
                         "reuse": reuse,
                         "variables_collections": collections}

  # Build the convolution neural network for this k-body atomic interaction.
  with arg_scope([conv2d],
                 weights_initializer=weights_initializer,
                 normalizer_fn=None,
                 reuse=reuse,
                 variables_collections=collections):
    for i, num_kernels in enumerate(num_kernels):
      with tf.variable_scope("Hidden{:d}".format(i + 1), reuse=reuse):
        inputs = conv2d(inputs,
                        kernel_size=kernel_size,
                        biases_initializer=biases_initializer,
                        num_outputs=num_kernels,
                        activation_fn=None,
                        trainable=trainable,
                        scope="1x1Conv{:d}".format(i + 1))
        if normalizer_fn is not None:
          inputs = normalizer_fn(inputs, **normalizer_params)
        inputs = activation_fn(inputs)
        if verbose:
          print_activations(inputs)

    outputs = conv2d(inputs,
                     kernel_size=kernel_size,
                     num_outputs=1,
                     biases_initializer=None,
                     activation_fn=None,
                     trainable=trainable,
                     scope="k-Body")
    if verbose:
      print_activations(outputs)
      tf.logging.info("")
    return outputs


def _inference_1body_nn(occurs, num_atom_types, initial_one_body_weights=None,
                        reuse=False, trainable=True):
  """
  Inference the one-body part.

  Args:
    occurs: a Tensor of shape `[-1, 1, 1, num_atom_types]` as the number of
      occurances of each type of atom.
    num_atom_types: a `int` as the number of atom types.
    initial_one_body_weights: an array of shape `[num_atom_types, ]` as the
      initial weights of the one-body kernel.
    reuse: a `bool` indicating whether we should reuse the variables or not.
    trainable: a `bool` indicating whether the one-body parameters are trainable
      or not.

  Returns:
    one_body: a 4D Tensor of shape `[-1, 1, 1, 1]` as the one-body contribs.

  """
  num_outputs = 1
  kernel_size = 1

  if initial_one_body_weights is not None:
    values = np.asarray(initial_one_body_weights, dtype=np.float32)
  else:
    values = np.ones(num_atom_types, dtype=np.float32)
  if len(values) != num_atom_types:
    raise Exception("The number of weights should be %d!" % num_atom_types)

  weights_initializer = init_ops.constant_initializer(values)
  return conv2d(
    occurs,
    num_outputs=num_outputs,
    kernel_size=kernel_size,
    activation_fn=None,
    weights_initializer=weights_initializer,
    biases_initializer=None,
    scope='one-body',
    reuse=reuse,
    variables_collections=[KcnnGraphKeys.ENERGY_VARIABLES],
    trainable=trainable,
  )


def inference_forces(y_total, inputs, coefficients, indexing, summary=True):
  """
  Inference the kCON forces.

  Args:
    y_total: a `float32` Tensor of shape `[-1, ]` as the output.
    inputs: a Tensor of shape `[-1, 1, D, C(k, 2)]` as the inputs.
    coefficients: a 3D Tensor as the auxiliary coefficients for computing atomic
      forces.
    indexing: a 3D Tensor as the indexing matrix for force compoenents.
    summary: a `bool` indicating whether we add summaries of some internal
      tensors or not.

  Returns:
    forces: a `float32` Tensor of shape `[-1, num_force_components]` as the
      neural network forces.

  """

  assert isinstance(coefficients, tf.Tensor)
  assert isinstance(indexing, tf.Tensor)

  with tf.name_scope("Forces"):

    # Compute the derivative of dE / dz. `z` is the input feature matrix and
    # E = NN(z) is the output of the KCNN model.
    dydz = tf.gradients([y_total], [inputs], name="dydz")[0]
    if summary:
      tf.summary.histogram(dydz.op.name + "/hist", dydz)

    # Squeeze the `dydz`. Now its shape will be `[-1, D, C(k, 2)]`
    dydz = tf.squeeze(dydz, axis=1, name="squeezed")

    # Tile the derivatives because each entry of `z` contributes to six force
    # components.
    tiled = tf.tile(dydz, (1, 1, 6), "tiled")

    # Do the element-wise multiplication. Now we get dE/dz * dz/dr * dr/d{x,y,z}
    # `g` has the shape of `[-1, D, 6 * C(k, 2)]`.
    g = tf.multiply(tiled, coefficients, name="g")

    # Now we should re-order all entries of `g`. Flatten it so that its shape
    # will be `[-1, D * 6 * C(k, 2)]`.
    g = flatten(g)

    # Pad an zero at the begining of each row of `g` because real indices in
    # `indexing` start from one and the index of zero suggests the contribution
    # should also be zero.
    g = tf.pad(g, [[0, 0], [1, 0]], mode='constant', name="pad")

    # The basic idea of the re-ordering algorithm is taking advantage of the
    # array broadcasting scheme of TensorFlow (Numpy). Since the batch size (the
    # first axis of `g`) will not be 1, we cannot do broadcasting directly.
    # Instead, we make the `g` a total flatten vector and broadcast it into a
    # matrix with `indexing`.
    with tf.name_scope("reshape"):

      with tf.name_scope("indices"):

        with tf.name_scope("g"):
          shape = tf.shape(g, name="shape")
          batch_size, step = shape[0], shape[1]

        with tf.name_scope("indexing"):
          shape = tf.shape(indexing, name="shape")
          num_f, num_entries = shape[1], shape[2]

        multiples = [1, num_f, num_entries]
        size = tf.multiply(batch_size, step, name="total_size")
        steps = tf.range(0, limit=size, delta=step, name="arange")
        steps = tf.reshape(steps, (batch_size, 1, 1), name="steps")
        steps = tf.tile(steps, multiples, name="tiled")
        indices = tf.add(indexing, steps, name="indices")

      # Do the broadcast
      g = tf.reshape(g, (-1, ), "1D")
      g = tf.gather(g, indices, name="gather")

      # Reshape `g` so that all entries of each row (axis=2) correspond to the
      # same force component (axis=1).
      g = tf.reshape(g, (batch_size, num_f, num_entries), "reshape")
      if summary:
        tf.summary.histogram(g.op.name + "/hist", g)

    # Sum up all entries of each row to get the final gradient for each force
    # component.
    g = tf.reduce_sum(g, axis=2, keep_dims=False, name="sum")

    # Always remember the physics law: f = -dE / dr. But the output `y_total`
    # already took the minus sign.
    forces = tf.identity(g, "forces")

  return forces


def _split_inputs(inputs, split_dims):
  """
  Split the inputs into different parts. Each part represents a k-body atomic
  interaction.
  """
  with tf.name_scope("Split"):
    axis = tf.constant(2, dtype=tf.int32, name="major")
    return tf.split(inputs, split_dims, axis=axis)


def inference_energy(inputs, occurs, weights, split_dims, num_atom_types,
                     kbody_terms, is_training, reuse=False, verbose=True,
                     num_kernels=None, activation_fn=lrelu,
                     normalizer='bias', weights_initializer=None,
                     one_body_weights=None, trainable_one_body=True,
                     trainable_k_max=3, summary=True):
  """
  Inference the kCON energy model.

  Args:
    inputs: a Tensor of shape `[-1, 1, -1, D]` as the inputs.
    occurs: a Tensor of shape `[-1, num_atom_types]` as the number of occurances
      of each type of atom.
    weights: a Tensor of shape `[-1, -1, D, -1]` as the weights of the k-body
      contribs.
    split_dims: a `List[int]` or a 1D Tensor of `int` as the dims to split the
      input feature matrix.
    num_atom_types: a `int` as the number of atom types.
    kbody_terms: a `List[str]` as the names of the k-body terms.
    is_training: a `bool` type placeholder indicating whether this inference is
      for training or not.
    reuse: a `bool` indicating whether we should reuse variables or not.
    verbose: boolean indicating whether the layers shall be printed or not.
    num_kernels: a `Tuple[int]` as the number of kernels of the convolution
      layers. This also determines the number of layers in each atomic network.
    activation_fn: a `Callable` as the activation function.
    normalizer: a `str` as the normalizer to use. Supported normalizers include:
      'bias' (default), 'batch_norm', 'layer_norm' and None.
    weights_initializer: a `Callable` as the function to initialize weights.
    one_body_weights: a `float32` array of shape `[num_atom_types, ]` as the
      initial weights of the one-body kernel.
    trainable_one_body: a `bool` indicating whether the one body parameters are
      trainable or not.
    trainable_k_max: an `int`. All k-body terms (2 <= k <= trainable_k_max) will
      be trained.
    summary: a `bool` indicating whether we should add summaries for
      tensors or not.

  Returns:
    y_total: a `float32` Tensor of shape `[-1, ]` as the total energies.
    y_contribs: a `float32` Tensor of shape `[-1, D]` as the predicted energies
      of the kbody contribs.
    forces: a `float32` Tensor of shape `[-1, num_force_components]` as the
      neural network forces.

  """

  with tf.name_scope("Energy"):

    # Split the input feature matrix into several parts. Each part represents a
    # certain atomic interaction. The number of parts is equal to the number of
    # k-body terms.
    splited_inputs = _split_inputs(inputs, split_dims)

    # Inference the convolution network for each k-body interaction
    y_contribs = []
    for i, conv in enumerate(splited_inputs):
      with tf.variable_scope(kbody_terms[i]):

        # Get the number of GHOST atoms in this k-body term so that we can
        # determine if this k-body CNN should be trained.
        symbols = get_atoms_from_kbody_term(kbody_terms[i])
        k = len(symbols) - symbols.count(GHOST)
        trainable = bool(k <= trainable_k_max)

        # For 2-body terms only the 1st column of the last dimension is needed.
        if k == 2:
          conv = conv[..., 0:1]

        # A gate tensor should be added so that non-trainable CNNs will not give
        # energy contributions.
        if trainable:
          gate = tf.constant(1.0, dtype=tf.float32, name='gate_on')
        else:
          gate = tf.constant(0.0, dtype=tf.float32, name="gate_off")

        # Do the inference
        y_contrib = _inference_kbody_cnn(
          inputs=conv,
          kbody_term=kbody_terms[i],
          activation_fn=activation_fn,
          reuse=reuse,
          normalizer=normalizer,
          weights_initializer=weights_initializer,
          is_training=is_training,
          num_kernels=num_kernels,
          trainable=trainable,
          verbose=verbose
        )
        y_gated = tf.multiply(y_contrib, gate, name='y_gated')
        y_contribs.append(y_gated)

    # Concat the k-body contribs from all k-body terms. The new tensor has the
    # shape of `[-1, 1, D, 1]`.
    contribs = tf.concat(y_contribs, axis=2, name="raw_contribs")

    # Obtain the weighted k-body contribs.
    # In general we hope zero inputs lead to zero contribs. But the convolution
    # kernels have biases so the output may not be zero. To fix this potential
    # problem we multiply the calculated k-body contribs with binary weights.
    contribs = tf.multiply(contribs, weights, name="y_contribs")
    if summary:
      tf.summary.histogram("kbody_contribs", contribs)
    if verbose:
      print_activations(contribs)

    # Inference the one-body expression.
    one_body = _inference_1body_nn(occurs,
                                   num_atom_types,
                                   reuse=reuse,
                                   initial_one_body_weights=one_body_weights,
                                   trainable=trainable_one_body)
    if summary:
      tf.summary.histogram("1body_contribs", one_body)
    if verbose:
      print_activations(one_body)

    # Sum up the k-body contribs and one-body contribs to get the total energy.
    # This is why we call this network `KCNN`.
    with tf.name_scope("Sum"):
      with tf.name_scope("kbody"):
        y_total_kbody = tf.reduce_sum(contribs, axis=2, name="total")
        y_total_kbody.set_shape([None, 1, 1])
        if verbose:
          print_activations(y_total_kbody)
        y_total_kbody = tf.squeeze(flatten(y_total_kbody), name="squeeze")
        if summary:
          tf.summary.scalar("Ek", tf.reduce_mean(y_total_kbody, name="avgEk"))
      with tf.name_scope("1body"):
        y_total_1body = tf.squeeze(one_body, name="squeeze")
        if summary:
          tf.summary.scalar(
            'logE1', tf.log(
              tf.reduce_mean(y_total_1body, name="avgE1"), name="log"))
      y_total = tf.add(y_total_1body, y_total_kbody, "1_and_k")

    if verbose:
      get_number_of_trainable_parameters(verbose=verbose)
    return y_total, contribs


def print_activations(tensor):
  """
  Print the name and shape of the input Tensor.

  Args:
    tensor: a Tensor.

  """
  dims = ",".join(["{:6d}".format(dim if dim is not None else -1)
                   for dim in tensor.get_shape().as_list()])
  tf.logging.info("%-36s : [%s]" % (tensor.op.name, dims))


def get_number_of_trainable_parameters(verbose=False):
  """
  Return the number of trainable parameters in current graph.

  Args:
    verbose: a bool. If True, the number of parameters for each variable will
    also be printed.

  Returns:
    ntotal: the total number of trainable parameters.

  """
  tf.logging.info("")
  tf.logging.info("Compute the total number of trainable parameters ...")
  tf.logging.info("")
  ntotal = 0
  for var in tf.trainable_variables():
    nvar = np.prod(var.get_shape().as_list(), dtype=int)
    if verbose:
      tf.logging.info("{:<38s}   {:>8d}".format(var.name, nvar))
    ntotal += nvar
  tf.logging.info("Total number of parameters: %d" % ntotal)
  tf.logging.info("")


def variable_summaries(tensor):
  """
  Attach a lot of summaries to a Tensor (for TensorBoard visualization).

  Args:
    tensor: a Tensor.

  """
  with tf.name_scope('summaries'):
    mean = tf.reduce_mean(tensor)
    tf.summary.scalar('mean', mean)
    with tf.name_scope('stddev'):
      stddev = tf.sqrt(tf.reduce_mean(tf.square(tf.subtract(tensor, mean))))
      tf.summary.scalar('stddev', stddev)
    tf.summary.scalar('max', tf.reduce_max(tensor))
    tf.summary.scalar('min', tf.reduce_min(tensor))
    tf.summary.histogram('histogram', tensor)


def debug():
  """
  Debug the KCNN inference.
  """
  properties = dict(
    num_atom_types=4,
    split_dims=[84, 252, 36, 36, 189, 63, 63, 9, 35, 21, 21, 7],
    atomic_forces=True,
    kbody_terms=["CCC", "CCH", "CCN", "CCX", "CHH", "CHN", "CHX", "CNX", "HHH",
                 "HHN", "HHX", "HNX"],
    one_body_weights=np.ones(4, dtype=np.float32),
  )

  graph = tf.Graph()

  with graph.as_default():

    inputs = tf.placeholder(
      tf.float32, shape=[50, 1, None, 3], name="inputs"
    )
    occurs = tf.placeholder(
      tf.float32, shape=[50, 1, 1, properties['num_atom_types']], name="occurs"
    )
    binary_weights = tf.placeholder(
      tf.float32, shape=[50, 1, None, 1], name="weights"
    )
    split_dims = tf.placeholder(
      tf.int64, shape=[len(properties['split_dims']), 0], name="split_dims"
    )
    coefficients = tf.placeholder(
      tf.float32, shape=[50, None, 18], name="aux_coef"
    )
    indexing = tf.placeholder(
      tf.int32, shape=[50, None, None], name="indexing"
    )

    y_total, y_contribs = inference_energy(
      inputs=inputs,
      occurs=occurs,
      weights=binary_weights,
      split_dims=split_dims,
      num_atom_types=properties['num_atom_types'],
      kbody_terms=properties['kbody_terms'],
      is_training=True,
      verbose=True,
      one_body_weights=properties['one_body_weights'],
    )

    f_calc = inference_forces(
      y_total=y_total,
      inputs=inputs,
      coefficients=coefficients,
      indexing=indexing
    )

    print(y_total.get_shape())
    print(y_contribs.get_shape())
    print(f_calc.get_shape())


if __name__ == "__main__":
  debug()



# coding=utf-8
"""
This script is used to transform atomic coordinates to input features and then
save them in tfrecords files.
"""

from __future__ import print_function, absolute_import

import numpy as np
import tensorflow as tf
import json
import sys
import time
from collections import Counter, namedtuple
from itertools import combinations, product, repeat, chain
from functools import partial
from os.path import basename, dirname, join, splitext
from ase.atoms import Atoms
from scipy.special import comb
from sklearn.metrics import pairwise_distances
from tensorflow.python.training.training import Features, Example
from .constants import pyykko, GHOST, LJR
from .utils import get_atoms_from_kbody_term, safe_divide, compute_n_from_cnk
from .utils import Gauss

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


# TODO: fix the multi-counting problem when the number of ghost atoms >= 2.
# TODO: implement `compute_atomic_energies` for `FixedLenMultiTransformer`.

FLAGS = tf.app.flags.FLAGS

# Important: this `_safe_log` must be very very small. The previous e^-6 is
# large enough to cause significant numeric errors.
_safe_log = np.e**(-20)


"""
A data structure for storing transformed features and auxiliary parameters.
"""
KcnnSample = namedtuple("KcnnSample", (
  "features",
  "split_dims",
  "binary_weights",
  "occurs",
  "coefficients",
  "indexing",
  "compress_stats"
))


def get_formula(species):
  """
  Return the molecular formula given a list of atomic species.
  """
  return "".join(species)


def _compute_lr_weights(coef, y, num_real_atom_types, factor=1.0):
  """
  Solve the linear equation system of Ax = b.
  
  Args:
    coef: a `float` array of shape `[num_examples, num_atom_types]`.
    y: a `float` array of shape `[num_examples, ]`.
    num_real_atom_types: an `int` as the number of atom types excluding the
      ghost atoms.
    factor: a `float` as a scaling factor for the weights.

  Returns:
    x: a `float` array of shape `[num_atom_types, ]` as the solution.

  """
  rank = np.linalg.matrix_rank(coef[:, :num_real_atom_types])
  diff = num_real_atom_types - rank

  # The coef matrix is full rank. So the linear equation system can be solved.
  if diff == 0:
    x = np.negative(np.dot(np.linalg.pinv(coef), y))

  # The rank is 1, so all structures have the same stoichiometry. Then all types
  # of atoms can be treated equally.
  elif rank == 1:
    x = np.negative(np.mean(y / coef[:, :num_real_atom_types].sum(axis=1)))

  else:
    raise ValueError(
      "Coefficients matrix rank {} of {} is not supported!".format(
        rank, num_real_atom_types))

  return x * factor


def _get_pyykko_bonds_matrix(species, factor=1.0, flatten=True, lj=False):
  """
  Return the pyykko-bonds matrix given a list of atomic symbols.

  Args:
    species: a `List[str]` as the atomic symbols.
    factor: a `float` as the normalization factor.
    flatten: a `bool` indicating whether the bonds matrix is flatten or not.
    lj: a `bool`. If True, all atoms will be treated as ideal LJ atoms.

  Returns:
    bonds: the bonds matrix (or vector if `flatten` is True).

  """
  if not lj:
    rr = np.asarray([pyykko[specie] for specie in species])[:, np.newaxis]
  else:
    rr = np.ones((len(species), 1)) * LJR
  lmat = np.multiply(factor, rr + rr.T)
  if flatten:
    return lmat.flatten()
  else:
    return lmat


def get_kbody_terms_from_species(species, k_max):
  """
  Return the k-body terms given the chemical symbols and `many_body_k`.

  Args:
    species: a `list` of `str` as the chemical symbols.
    k_max: a `int` as the maximum k-body terms that we should consider.

  Returns:
    kbody_terms: a `list` of `str` as the k-body terms.

  """
  return sorted(list(set(
      ["".join(sorted(c)) for c in combinations(species, k_max)])))


def _get_num_force_entries(n, k_max):
  """
  Return the number of entries per force component.

  Args:
    n: an `int` as the maximum number of 'real' atoms in a structure.
    k_max: an `int` as the maximum k.

  Returns:
    num_entries: an `int` as the number of entries per force component.

  """
  return int(np.array([comb(n, k) * comb(k, 2) * 2 / n
                       for k in range(2, k_max + 1)]).sum())


def exponential_norm(x, unit=1.0, order=1):
  """
  Normalize the inputs `x` with the exponential function:
    f(x) = exp(-x / unit)

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).
    order: a `int` as the exponential order. If `order` is 0, the inputs will 
      not be scaled by `factor`.

  Returns:
    scaled: the scaled unitless inputs.

  """
  order = int(order)
  if order == 0:
    return np.exp(-x)
  else:
    return np.exp(-(x / unit) ** order)


def exponential_norm_prime(x, unit=1.0, order=1, z=None):
  """
  The derivative of the exponential norm function.

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).
    order: a `int` as the exponential order. If `order` is 0, the inputs will
      not be scaled by `factor`.
    z: the pre-computed value of `exponential_norm(x, unit, order)`.

  Returns:
    g: the gradient.

  """
  order = int(order)
  if z is None:
    z = exponential_norm(x, unit=unit, order=order)
  if order == 0:
    return -z
  if order == 1:
    return -z / unit
  else:
    return z * (-order * x**(order - 1) / unit**order)


def exponential_gauss(x, unit=1.0):
  """
  Normalize the inputs `x` with the mixed functions:
    f(x) = exp(-x / unit),               if x / unit >= 1.0
    f(x) = gaussian(1.0, 0.2)(x / unit), else

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).

  Returns:
    scaled: the scaled unitless inputs.

  """
  g = Gauss(1.0, 0.2)
  r = x / unit
  scale = np.exp(-1.0) / g(1.0)
  if hasattr(r, '__getitem__'):
    left = np.where(r < 1.0)[0]
    y = np.exp(-r)
    y[left] = g(r[left]) * scale
    return y
  else:
    if r < 1.0:
      return g(r) * scale
    else:
      return np.exp(-r)


def lj_norm(x, unit=1.0):
  """
  Normalize the inputs `x` with the Lennard-Jones function.

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).

  Returns:
    scaled: the scaled unitless inputs.

  """
  rm = x / unit
  left = np.where(rm < 1.0)[0]
  rp = rm ** 4
  rp[left] = rm[left] ** 2
  y = safe_divide(1.0, rp)
  return np.clip(-(y ** 2 - 2.0 * y), 0.0, 1.0)


def morse_norm(x, unit=1.0, order=1):
  """
  Normalize the inputs `x` with the Morse function.

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).
    order: a `float` or an `int`.

  Returns:
    scaled: the scaled unitless inputs.

  """
  alpha = -1.0 * float(order)
  return (1.0 - np.exp(alpha * (x - unit)))**2 - np.exp(-1.0)


# noinspection PyUnusedLocal
def morse_norm_prime(x, unit=1.0, order=1, z=None):
  """
  The derivative of the `morse` function.

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).
    order: a `float` or an `int`.
    z: a dummy variable.

  Returns:
    g: the gradient.

  """
  alpha = -1.0 * float(order)
  g = np.exp(alpha * (x - unit))
  return 2.0 * g * alpha * (g - 1.0)


# noinspection PyUnusedLocal
def inverse_norm(x, unit=1.0, order=1):
  """
  The simple inverse function: z(x) = 1.0 / x

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a dummy variable.
    order: a dummy variable.

  Returns:
    z: the inversed inputs.

  """
  return 1.0 / x


# noinspection PyUnusedLocal
def inverse_norm_prime(x, unit=1.0, order=1, z=None):
  """
  The derivative of the `inverse_norm` function.

  Args:
    x: Union[float, np.ndarray] as the inputs to scale.
    unit: a `float` or an array with the same shape of `inputs` as the scaling
      factor(s).
    order: a `float` or an `int`.
    z: a dummy variable.

  Returns:
    g: the gradient.

  """
  if z is None:
    z = inverse_norm(x)
  return -z**2


def _bytes_feature(value):
  """
  Convert the `value` to Protobuf bytes.
  """
  return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _float_feature(value):
  """
  Convert the `value` to Protobuf float32.
  """
  return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


class Transformer:
  """
  This class is used to transform atomic coordinates to input feature matrix.
  """

  def __init__(self, species, k_max=3, kbody_terms=None, split_dims=None,
               norm='exp', norm_order=1, periodic=False, atomic_forces=False,
               lj=False, cutoff=None):
    """
    Initialization method.

    Args:
      species: a `List[str]` as the ordered atomic symboles.
      k_max: a `int` as the maximum k for the many body expansion scheme.
      kbody_terms: a `List[str]` as the k-body terms.
      split_dims: a `List[int]` as the dimensions for splitting inputs. If this
        is given, the `kbody_terms` must also be set and their lengths should be
        equal.
      norm: a `str` specifying the normalization function to use. Defaults to
        `exp`, `lj`, `exp+g`, 'inv' and `morse` are also supported.
      norm_order: a `int` as the order for normalizing interatomic distances.
      periodic: a `bool` indicating whether this transformer is used for 
        periodic structures or not.
      atomic_forces: a `bool` indicating whether the atomic forces derivation is
        enabled or not.
      lj: a `bool` indicating that this transformer targets on LJ systems.
      cutoff: a `float` as the cutoff.

    """
    if split_dims is not None:
      assert len(split_dims) == len(kbody_terms)

    kbody_terms = kbody_terms or get_kbody_terms_from_species(species, k_max)
    num_ghosts = self._get_num_ghosts(species, k_max)
    mapping, selections = self._get_mapping(species, kbody_terms)

    # Internal initialization.
    offsets, real_dim, kbody_sizes = [0], 0, []
    if split_dims is None:
      # If `split_dims` is not given, we shall construct it by our own.
      # `real_dim` should be always not smaller than `sum(kbody_sizes)` because
      # every excluded k-body term is represented by a single row vector of all
      # zeros.
      for kbody_term in kbody_terms:
        if kbody_term in mapping:
          size = mapping[kbody_term].shape[1]
        else:
          size = 0
        real_dim += max(size, 1)
        kbody_sizes.append(size)
        offsets.append(real_dim)
      split_dims = np.diff(offsets).tolist()
      # Here we must use `sum(kbody_sizes)` but not `offsets[-1]` because when
      # `split_dims` is not set while `kbody_terms` is given, `offsets[-1]` is
      # not equal to C(N_max, k).
      n = compute_n_from_cnk(sum(kbody_sizes), k_max)

    else:
      offsets = [0] + np.cumsum(split_dims).tolist()
      real_dim = offsets[-1]
      kbody_sizes = []
      for kbody_term in kbody_terms:
        if kbody_term in mapping:
          size = mapping[kbody_term].shape[1]
        else:
          size = 0
        kbody_sizes.append(size)
      n = compute_n_from_cnk(offsets[-1], k_max)

    # Initialize internal variables.
    self._lj = lj
    self._k_max = k_max
    self._kbody_terms = kbody_terms
    self._offsets = offsets
    self._kbody_sizes = kbody_sizes
    self._species = species
    self._mapping = mapping
    self._selections = selections
    self._split_dims = split_dims
    self._ck2 = int(comb(k_max, 2, exact=True))
    self._cond_sort = self._get_conditional_sorting_indices(kbody_terms)
    self._cmatrix = _get_pyykko_bonds_matrix(species, lj=lj)
    self._num_ghosts = num_ghosts
    self._periodic = periodic
    self._real_dim = real_dim
    self._binary_weights = self._get_binary_weights()
    self._atomic_forces = atomic_forces
    self._indexing_matrix = None
    self._num_real = n - self._num_ghosts
    self._num_f_components = 3 * self._num_real
    self._num_entries = _get_num_force_entries(self._num_real, self._k_max)
    self._norm = norm
    self._norm_order = norm_order
    if norm.lower() == 'exp':
      self._norm_fn = partial(exponential_norm, order=norm_order)
      self._norm_prime_fn = partial(exponential_norm_prime, order=norm_order)
    elif norm.lower() == 'lj':
      self._norm_fn = lj_norm
      self._norm_prime_fn = None
    elif norm.lower() == 'exp+g':
      self._norm_fn = exponential_gauss
      self._norm_prime_fn = None
    elif norm.lower() == 'morse':
      self._norm_fn = partial(morse_norm, order=norm_order)
      self._norm_prime_fn = partial(morse_norm_prime, order=norm_order)
    elif norm.lower() == 'inv':
      self._norm_fn = inverse_norm
      self._norm_prime_fn = inverse_norm_prime
    else:
      raise ValueError("Unsupported normalizing function: {}".format(norm))
    self._cutoff = cutoff or np.inf
    self._cutoff_table = self._get_cutoff_table()

  @property
  def species(self):
    """
    Return the species of this transformer excluding all ghosts.
    """
    return [symbol for symbol in self._species if symbol != GHOST]

  @property
  def shape(self):
    """
    Return the shape of the transformed input feature matrix.
    """
    return self._real_dim, self._ck2

  @property
  def ck2(self):
    """
    Return the value of C(k,2) for this transformer.
    """
    return self._ck2

  @property
  def k_max(self):
    """
    Return the maximum order for the many-body expansion.
    """
    return self._k_max

  @property
  def split_dims(self):
    """
    Return the dims for spliting the inputs.
    """
    return self._split_dims

  @property
  def kbody_terms(self):
    """
    Return the kbody terms for this transformer. 
    """
    return self._kbody_terms

  @property
  def kbody_sizes(self):
    """
    Return the real sizes of each kbody term of this transformer. Typically this
    is equal to `split_dims` but when `kbody_terms` is manually set, this may be 
    different.
    """
    return self._kbody_sizes

  @property
  def binary_weights(self):
    """
    Return the binary weights for the all k-body contribs.
    """
    return self._binary_weights

  @property
  def kbody_selections(self):
    """
    Return the kbody selections.
    """
    return self._selections

  @property
  def num_ghosts(self):
    """
    Return the number of ghosts atoms.
    """
    return self._num_ghosts

  @property
  def is_periodic(self):
    """
    Return True if this transformer is used for periodic structures.
    """
    return self._periodic

  @property
  def support_atomic_forces(self):
    """
    Return True if atomic forces derivation is enabled.
    """
    return self._atomic_forces

  @property
  def num_force_components(self):
    """
    Return the total number of force components.
    """
    return self._num_f_components

  @property
  def num_entries_per_component(self):
    """
    Return the number of entries per force component.
    """
    return self._num_entries

  @property
  def is_lj(self):
    """
    Return True if this transformer shall be used for LJ systems.
    """
    return self._lj

  @property
  def cutoff(self):
    """
    Return the cutoff.
    """
    return self._cutoff

  def get_bond_types(self):
    """
    Return the ordered bond types for each k-body term.
    """
    bonds = {}
    for kbody_term in self._kbody_terms:
      atoms = get_atoms_from_kbody_term(kbody_term)
      bonds[kbody_term] = ["-".join(ab) for ab in combinations(atoms, r=2)]
    return bonds

  def _get_cutoff_table(self):
    """
    Return the cutoff table.
    """
    table = np.zeros((self._real_dim, self._ck2), dtype=np.float32)
    cutoff = self._cutoff
    if cutoff < np.inf:
      thres = self._norm_fn(cutoff)
    else:
      thres = 0.0
    for i, kbody_term in enumerate(self._kbody_terms):
      istart, istop = self._offsets[i], self._offsets[i + 1]
      if GHOST in kbody_term:
        table[istart: istop, 0] = thres
      else:
        table[istart: istop, :] = thres
    return table

  @staticmethod
  def _get_num_ghosts(species, many_body_k):
    """
    Return and check the number of ghost atoms.

    Args:
      species: a `list` of `str` as the chemical symbols.
      many_body_k: a `int` as the maximum k-body terms that we should consider.

    Returns:
      num_ghosts: a `int` as the number of ghost atoms.

    """
    num_ghosts = list(species).count(GHOST)
    if num_ghosts != 0 and (num_ghosts > 2 or many_body_k - num_ghosts != 2):
      raise ValueError("The number of ghosts is wrong!")
    return num_ghosts

  @staticmethod
  def _get_mapping(species, kbody_terms):
    """
    Build the mapping from interatomic distance matrix of shape `[N, N]` to the
    input feature matrix of shape `[C(N, k), C(k, 2)]`.

    Args:
      species: a `list` of `str` as the ordered atomic symbols.
      kbody_terms: a `list` of `str` as the ordered k-body terms.

    Returns:
      mapping: a `Dict[str, Array]` as the mapping from the N-by-N interatomic
        distance matrix to the input feature matrix for each k-body term.
      selection: a `Dict[str, List[List[int]]]` as the indices of the k-atoms
        selections for each k-body term.

    """
    natoms = len(species)
    mapping = {}
    selections = {}

    # Determine the indices of each type of atom and store them in a dict.
    atom_index = {}
    for i in range(len(species)):
      atom = species[i]
      atom_index[atom] = atom_index.get(atom, []) + [i]

    for kbody_term in kbody_terms:
      # Extract atoms from this k-body term
      atoms = get_atoms_from_kbody_term(kbody_term)
      # Count the occurances of the atoms.
      counter = Counter(atoms)
      # If an atom appears more in the k-body term, we should discard this
      # k-body term. For example the `CH4` molecule can not have `CCC` or `CCH`
      # interactions.
      if any([counter[e] > len(atom_index.get(e, [])) for e in atoms]):
        continue
      # ck2 is the number of bond types in this k-body term.
      ck2 = int(comb(len(atoms), 2, exact=True))
      # Sort the atoms
      sorted_atoms = sorted(counter.keys())
      # Build up the k-atoms selection candidates. For each type of atom we draw
      # N times where N is equal to `counter[atom]`. Thus, the candidate list
      # can be constructed:
      # [[[1, 2], [1, 3], [1, 4], ...], [[8], [9], [10], ...]]
      # The length of the candidates is equal to the number of atom types.
      k_atoms_candidates = [
        [list(o) for o in combinations(atom_index[e], counter[e])]
        for e in sorted_atoms
      ]
      # Build up the k-atoms selections. First, we get the `product` (See Python
      # official document for more info), eg [[1, 2], [8]]. Then `chain` it to
      # get flatten lists, eg [[1, 2, 8]].
      k_atoms_selections = [list(chain(*o)) for o in
                            product(*k_atoms_candidates)]
      selections[kbody_term] = k_atoms_selections
      # cnk is the number of k-atoms selections.
      cnk = len(k_atoms_selections)
      # Construct the mapping from the interatomic distance matrix to the input
      # matrix. This procedure can greatly increase the transformation speed.
      # The basic idea is to fill the input feature matrix with broadcasting.
      # The N-by-N interatomic distance matrix is flatten to 1D vector. Then we
      # can fill the matrix like this:
      #   feature_matrix[:, col] = flatten_dist[[1,2,8,10,9,2,1,1]]
      mapping[kbody_term] = np.zeros((ck2, cnk), dtype=int)
      for i in range(cnk):
        for j, (vi, vj) in enumerate(combinations(k_atoms_selections[i], 2)):
          mapping[kbody_term][j, i] = vi * natoms + vj
    return mapping, selections

  @staticmethod
  def _get_conditional_sorting_indices(kbody_terms):
    """
    Generate the indices of the columns for the conditional sorting scheme.

    Args:
      kbody_terms: a `List[str]` as the ordered k-body terms.

    Returns:
      indices: a `dict` of indices for sorting along the last axis of the input
        features.

    """
    indices = {}
    for kbody_term in kbody_terms:
      # Extract atoms from this k-body term
      atoms = get_atoms_from_kbody_term(kbody_term)
      # All possible bonds from the given atom types
      bonds = list(combinations(atoms, r=2))
      n = len(bonds)
      counter = Counter(bonds)
      # If the bonds are unique, there is no need to sort because columns of the
      # formed feature matrix will not be interchangable.
      if max(counter.values()) == 1:
        continue
      # Determine the indices of duplicated bonds.
      indices[kbody_term] = []
      for bond, times in counter.items():
        if times > 1:
          indices[kbody_term].append([i for i in range(n) if bonds[i] == bond])
    return indices

  def _get_binary_weights(self):
    """
    Return the binary weights.
    """
    weights = np.zeros(self._real_dim, dtype=np.float32)
    offsets = self._offsets
    for i in range(len(self._split_dims)):
      weights[offsets[i]: offsets[i] + self._kbody_sizes[i]] = 1.0
    return weights

  def _get_coords(self, atoms):
    """
    Return the N-by-3 coordinates matrix for the given `ase.Atoms`.

    Notes:
      Auxiliary vectors may be appended if `num_ghosts` is non-zero.

    """
    if self._num_ghosts > 0:
      # Append `num_ghosts` rows of zeros to the positions. We can not directly
      # use `inf` because `pairwise_distances` and `get_all_distances` do not
      # support `inf`.
      aux_vecs = np.zeros((self._num_ghosts, 3))
      coords = np.vstack((atoms.get_positions(), aux_vecs))
    else:
      coords = atoms.get_positions()
    return coords

  def _get_interatomic_distances(self, coords, cell, pbc):
    """
    Return the interatomic distances matrix and its associated coordinates
    differences matrices.

    Returns:
      dists: a `float32` array of shape `[N, N]` where N is the number of atoms
        as the interatomic distances matrix.
      delta: a `float32` array of shape `[N, N, 3]`. The last axis represents
        the X, Y, Z directions. `delta[:, :, 0]` is the pairwise distances of
        all X coordinates.

    """
    delta = None

    if not self.is_periodic:
      dists = pairwise_distances(coords)
      if self._atomic_forces:
        # Compute dx_ij, dy_ij and dz_ij. `sklearn.metrics.pairwise_distances`
        # cannot be used here because it will return absolute values.
        ndim = len(dists)
        delta = np.zeros((ndim, ndim, 3))

        # Based on my tests, the vector subtraction used here should be:
        # r_{ij} = r_{i} - r_{j}
        for i in range(ndim):
          for j in range(i + 1, ndim):
            delta[i, j, :] = coords[i] - coords[j]  # r_ij
            delta[j, i, :] = coords[j] - coords[i]  # r_ji
    else:
      atoms = Atoms(
        symbols=self._species,
        positions=coords,
        cell=cell,
        pbc=pbc
      )
      dists = atoms.get_all_distances(mic=True)
      if self._atomic_forces:
        raise ValueError("Atomic forces for periodic structures are not "
                         "supported yet!")
      del atoms

    # Manually set the distances between ghost atoms and real atoms to inf.
    if self._num_ghosts > 0:
      dists[:, -self._num_ghosts:] = np.inf
      dists[-self._num_ghosts:, :] = np.inf

      # Set the delta differences to 0.
      if self._atomic_forces:
        delta[:, -self._num_ghosts:, :] = 0.0
        delta[-self._num_ghosts:, :, :] = 0.0

    return dists, delta

  def _initialize_features(self, dists, delta, features=None):
    """
    Initialize the input feature matrix and build the auxiliary matrices.

    Args:
      dists: a `float32` array of shape `[N**2, ]` as the raw flatten
        interatomic distances matrix.
      delta: a `float32` array of shape `[N**2, 3]`.
      features: a 2D `float32` array or None as the location into which the
        result is stored. If not provided, a new array will be allocated.

    Returns:
      features: a `float32` array of shape `self.shape` as the input feature
        matrix.
      rr: a `float32` array of shape `self.shape` as the corresponding
        interatomic distances to `features`.
      cr: a `float32` array of shape `self.shape` as the covalent radii for
        corresponding entries.
      dr: a `floar32` array of shape `[self.shape[0], 6 * self.shape[1]]` as the
        corresponding differences of coordinates.

    """
    if features is None:
      features = np.zeros((self._real_dim, self._ck2))
    elif features.shape != self.shape:
      raise ValueError("The shape should be {}".format(self.shape))

    if self._atomic_forces:
      cr = np.zeros_like(features)
      dr = np.zeros((self._real_dim, self._ck2 * 6))
      rr = np.zeros_like(features)
    else:
      cr = None
      dr = None
      rr = None

    half = self._ck2 * 3
    step = self._ck2
    zero = 0

    # Transform the raw interatomic distances
    norm_dists = self._norm_fn(dists, unit=self._cmatrix.flatten())

    for i, kbody_term in enumerate(self._kbody_terms):
      if kbody_term not in self._mapping:
        continue
      # The index matrix was transposed because typically C(N, k) >> C(k, 2).
      # See `_get_mapping`.
      mapping = self._mapping[kbody_term]
      istart = self._offsets[i]
      # Manually adjust the step size because the offset length may be larger if
      # `split_dims` is fixed.
      istep = min(self._offsets[i + 1] - istart, mapping.shape[1])
      istop = istart + istep
      for k in range(self._ck2):
        features[istart: istop, k] = norm_dists[mapping[k]]
        if self._atomic_forces:
          cr[istart: istop, k] = self._cmatrix[mapping[k]]
          rr[istart: istop, k] = dists[mapping[k]]
          # x = 0, y = 1, z = 2
          for j in range(3):
            dr[istart: istop, j * step + k + zero] = +delta[mapping[k], j]
            dr[istart: istop, j * step + k + half] = -delta[mapping[k], j]

    return features, rr, cr, dr

  def _conditionally_sort(self, features, rr, cr, dr):
    """
    Apply the conditional sorting algorithm.

    Args:
      features: a `float32` array of shape `self.shape` as the input feature
        matrix.
      rr: a `float32` array of shape `self.shape` as the corresponding
        interatomic distances to `features`.
      cr: a `float32` array of shape `self.shape` as the covalent radii for
        corresponding entries.
      dr: a `floar32` array of shape `[self.shape[0], 6 * self.shape[1]]` as the
        corresponding differences of coordinates.

    Returns:
      features: a `float32` array of shape `self.shape` as the input feature
        matrix.
      rr: a `float32` array of shape `self.shape` as the corresponding
        interatomic distances to `features`.
      cr: a `float32` array of shape `self.shape` as the covalent radii for
        corresponding entries.
      dr: a `floar32` array of shape `[self.shape[0], 6 * self.shape[1]]` as the
        corresponding differences of coordinates.
      indexing: a `int` array of shape `[self.shape[0], self.shape[1], 2]`.

    """

    def cond_sort(_part, _orders, _shape):
      """
      Do the conditional sorting.
      """
      return _part.flatten()[_orders].reshape(_shape)

    indexing = self._get_indexing_matrix().copy()
    ck2 = self._ck2

    for i, kbody_term in enumerate(self._kbody_terms):

      if kbody_term not in self._mapping:
        continue

      for ix in self._cond_sort.get(kbody_term, []):
        z = features[self._offsets[i]: self._offsets[i + 1], ix]

        # The simple case: the derivation of atomic forces is disabled.
        if not self._atomic_forces:
          # `samples` is a 2D array, the Python advanced slicing will make the
          # returned `z` a copy but not a view. The shape of `z` is transposed.
          # So we should sort along axis 0 here!
          z.sort()
          features[self._offsets[i]: self._offsets[i + 1], ix] = z

        # When atomic forces are required, a different strategy should be used.
        # The two auxiliary matrices, covalent radii matrix and the coordinates
        # differences matrix, must all be sorted. So we get the sorting orders
        # at first.
        else:
          orders = np.argsort(z)
          istart = self._offsets[i]
          istop = self._offsets[i + 1]
          shape = z.shape
          step = shape[1]
          orders += np.tile(np.arange(0, z.size, step), (step, 1)).T

          # Sort the features
          features[istart: istop, ix] = cond_sort(z, orders, shape)

          # Sort the covalend radii
          cr[istart: istop, ix] = cond_sort(
            cr[istart: istop, ix], orders, shape)

          # Sort the interatomic distances
          rr[istart: istop, ix] = cond_sort(
            rr[istart: istop, ix], orders, shape)

          # Sort the indices
          for axis in range(2):
            array = indexing[istart: istop, :, axis]
            array[:, ix] = cond_sort(array[:, ix], orders, shape)
            indexing[istart: istop, :, axis] = array

          # Sort the delta coordinates: dx, dy, dz
          for axis in range(6):
            cstart = axis * ck2
            cstop = cstart + ck2
            array = dr[istart: istop, cstart: cstop]
            array[:, ix] = cond_sort(array[:, ix], orders, shape)
            dr[istart: istop, cstart: cstop] = array

    return features, rr, cr, dr, indexing


  def _get_coef_matrix(self, z, r, l, d6):
    """
    Return the tiled coefficients matrix `dE/dz * dz/d{px,py,pz}`.

    Args:
      z: a `float32` array of shape `self.shape` as the input feature matrix.
      r: a `float32` array of shape `self.shape` as the interatomic distances
        corresponding to `z`.
      l: a `float32` array of shape `self.shape` as the covalent radii matrix.
      d6: a `float32` array of shape `[self.shape[0], self.shape[1] * 6]` as the
        differences of the coordinates.

    Returns:
      coef: a `float32` array as the coefficients matrix. The shape of `coef` is
        the same with input `d6`.

    """
    if self._atomic_forces:
      if self._norm_prime_fn is None:
        raise ValueError("The normalization function `{}` does not support "
                         "predicting forces yet!".format(self._norm))
      g = self._norm_prime_fn(r, unit=l, z=z)
      r6 = np.tile(r, (1, 6))
      coef = np.tile(g, (1, 6)) * safe_divide(d6, r6)
      return np.nan_to_num(coef)

    else:
      return None

  def _get_indexing_matrix(self):
    """
    Return the matrix for indexing the gradients matrix.
    """
    if self._indexing_matrix is None:
      cnk = self._real_dim
      ck2 = self._ck2
      # Set all entries to -1
      index_matrix = np.zeros((cnk, ck2, 2), dtype=int) - 1
      for i, kbody_term in enumerate(self._kbody_terms):
        if kbody_term not in self._selections:
          continue
        selections = self._selections[kbody_term]
        offset = self._offsets[i]
        for j, selection in enumerate(selections):
          for l, (a, b) in enumerate(combinations(selection, r=2)):
            index_matrix[offset + j, l] = a, b
      self._indexing_matrix = index_matrix
    return self._indexing_matrix

  def _transform_indexing_matrix(self, indexing):
    """
    Transform the conditionally sorted indexing matrix.

    Args:
      indexing: an `int` array of shape `[self.shape[0], self.shape[1], 2]` as
        the indexing matrix.

    Returns:
      positions: an `int` array of shape `[3N, C(N, k) * C(k, 2) * 2 / N]` as
        the positions of the entries for each atomic force component.

    """
    if not self._atomic_forces:
      return None

    cnk = indexing.shape[0]
    ck2 = indexing.shape[1]
    positions = np.zeros((self._num_f_components, self._num_entries), dtype=int)
    loc = np.zeros((self._num_real, ), dtype=int)
    position = 0
    half = ck2 * 3
    zero = 0
    imax = len(self.species)

    # The index should start from 1. 0 will be used as the virtual index
    # corresponding to zero contribution.
    start = 1

    for i in range(cnk):
      if indexing[i].min() >= 0:
        for j in range(ck2):
          a, b = indexing[i, j, :]
          # The contributions from Atom-Ghost pairs should be ignored.
          if a >= imax or b >= imax:
            continue
          ax = a * 3 + 0
          ay = a * 3 + 1
          az = a * 3 + 2
          bx = b * 3 + 0
          by = b * 3 + 1
          bz = b * 3 + 2
          positions[ax, loc[a]] = position + 0 * ck2 + j + zero + start
          positions[ay, loc[a]] = position + 1 * ck2 + j + zero + start
          positions[az, loc[a]] = position + 2 * ck2 + j + zero + start
          loc[a] += 1
          positions[bx, loc[b]] = position + 0 * ck2 + j + half + start
          positions[by, loc[b]] = position + 1 * ck2 + j + half + start
          positions[bz, loc[b]] = position + 2 * ck2 + j + half + start
          loc[b] += 1

      position += 6 * ck2

    return positions

  def compress(self, features):
    """
    Apply the soft compressing algorithm. The compression is implemented by
    adjusting the binary weights.

    Args:
      features: a `float32` array of shape `self.shape` as the input feature
        matrix.

    Returns:
      weights: a `float32` array as the updated binary weights.
      counter: a `dict` to count the number of kept contribs for each k-body
        term. This may be an empty `dict` indicating all contribs are kept.

    """
    if self._cutoff == np.inf:
      return self._binary_weights, {}

    else:
      results = np.sum(features >= self._cutoff_table, axis=1, dtype=int)
      weights = np.ones_like(self._binary_weights)
      weights[results < 3] = 0.0
      counter = {}
      for i, kbody_term in enumerate(self._kbody_terms):
        istart, istop = self._offsets[i], self._offsets[i + 1]
        counter[kbody_term] = np.sum(results[istart: istop] == 3)
      return weights, counter

  def transform(self, atoms, features=None):
    """
    Transform the given `ase.Atoms` object to an input feature matrix.

    Args:
      atoms: an `ase.Atoms` object.
      features: a 2D `float32` array or None as the location into which the
        result is stored. If not provided, a new array will be allocated.

    Returns:
      features: a `float32` array of shape `self.shape` as the input feature
        matrix.
      coef: a `float32` array as the coefficients matrix. The shape of `coef` is
        `[self.shape[0], self.shape[1] * 6]`.
      indexing: an `int` array of shape `[3N, C(N, k) * C(k, 2) * 2 / N]` as the
        indices of the entries for each atomic force component.

    """

    # Get the coordinates matrix.
    coords = self._get_coords(atoms)

    # Compute the interatomic distances. For non-periodic molecules we use the
    # faster method `pairwise_distances`.
    dists, delta = self._get_interatomic_distances(
      coords,
      cell=atoms.get_cell(),
      pbc=atoms.get_pbc()
    )

    # Initialize the input feature matrix and auxiliary matrices.
    dists = dists.flatten()
    if delta is not None:
      delta = delta.reshape((-1, 3))
    features, rr, cr, dr = self._initialize_features(
      dists, delta, features=features)

    # Apply the conditional sorting algorithm
    features, rr, cr, dr, indexing = self._conditionally_sort(
      features, rr, cr, dr)

    # Get coefficients matrix for computing atomic forces.
    coef = self._get_coef_matrix(features, rr, cr, dr)

    # Transform the conditionally sorted indexing matrix.
    indexing = self._transform_indexing_matrix(indexing)

    # Convert the data types
    features = features.astype(np.float32)
    if self._atomic_forces:
      coef = coef.astype(np.float32)
      indexing = indexing.astype(np.int32)

    return features, coef, indexing


class MultiTransformer:
  """
  A flexible transformer targeting on AxByCz ... molecular compositions.
  """

  def __init__(self, atom_types, k_max=3, max_occurs=None, norm='exp',
               norm_order=1, include_all_k=True, periodic=False, lj=False,
               atomic_forces=False, cutoff=None):
    """
    Initialization method.

    Args:
      atom_types: a `List[str]` as the atomic species.
      k_max: a `int` as the maximum k for the many body expansion scheme.
      max_occurs: a `Dict[str, int]` as the maximum appearances for a specie. 
        If an atom is explicitly specied, it can appear infinity times.
      norm: a `str` specifying the normalization function to use. Defaults to
        `exp`, `lj`, `exp+g`, 'inv' and `morse` are also supported.
      norm_order: an `int` or a `float` as the order for normalizing interatomic
        distances.
      include_all_k: a `bool` indicating whether we shall include all k-body
        terms where k = 1, ..., k_max. If False, only k = 1 and k = k_max are
        considered.
      periodic: a `bool` indicating whether we shall use periodic boundary 
        conditions.
      atomic_forces: a `bool` indicating whether the atomic forces derivation is
        enabled or not.
      lj: a `bool` indicating that this transformer targets on LJ systems.
      cutoff: a `float` as the cutoff.

    """
    # Make sure the ghost atom is always the last one!
    if include_all_k and k_max == 3:
      num_ghosts = 1
      atom_types = list(atom_types)
      if GHOST in atom_types:
        if atom_types[-1] != GHOST:
          atom_types.remove(GHOST)
          atom_types = sorted(atom_types) + [GHOST]
        else:
          atom_types = sorted(atom_types[:-1]) + [GHOST]
      else:
        atom_types = sorted(atom_types) + [GHOST]
    elif k_max > 3:
      raise ValueError("k_max > 3 is not supported!")
    else:
      num_ghosts = 0
      if GHOST in atom_types:
        raise ValueError("GHOST is not allowed when k_max == 2!")

    # Determine the species and maximum occurs.
    species = []
    max_occurs = {} if max_occurs is None else dict(max_occurs)
    max_occurs[GHOST] = num_ghosts
    for specie in atom_types:
      species.extend(list(repeat(specie, max_occurs.get(specie, k_max))))
      if specie not in max_occurs:
        max_occurs[specie] = np.inf

    self._include_all_k = include_all_k
    self._k_max = k_max
    self._atom_types = atom_types
    self._species = species
    self._num_atom_types = len(atom_types)
    self._num_ghosts = num_ghosts
    self._kbody_terms = get_kbody_terms_from_species(species, k_max)
    self._transformers = {}
    self._max_occurs = max_occurs
    self._norm_order = norm_order
    self._periodic = periodic
    self._atomic_forces = atomic_forces
    self._lj = lj
    self._norm = norm
    self._cutoff = cutoff

    # The global split dims is None so that internal `_Transformer` objects will
    # construct their own `splid_dims`.
    self._split_dims = None

  @property
  def k_max(self):
    """
    Return the many-body expansion factor.
    """
    return self._k_max

  @property
  def included_k(self):
    """
    Return the included k.
    """
    if self._include_all_k:
      return list(range(1, self._k_max + 1))
    else:
      return [1, self._k_max]

  @property
  def ck2(self):
    """
    Return the value of C(k,2).
    """
    return comb(self._k_max, 2, exact=True)

  @property
  def kbody_terms(self):
    """
    Return the ordered k-body terms for this transformer.
    """
    return self._kbody_terms

  @property
  def species(self):
    """
    Return the ordered species of this transformer.
    """
    return self._species

  @property
  def atom_types(self):
    """
    Return the supported atom types.
    """
    return self._atom_types

  @property
  def number_of_atom_types(self):
    """
    Return the number of atom types in this transformer.
    """
    return self._num_atom_types

  @property
  def include_all_k(self):
    """
    Return True if a standalone two-body term is included.
    """
    return self._include_all_k

  @property
  def is_periodic(self):
    """
    Return True if this is a periodic transformer.
    """
    return self._periodic

  @property
  def max_occurs(self):
    """
    Return the maximum occurances of each type of atom.
    """
    return self._max_occurs

  @property
  def atomic_forces_enabled(self):
    """
    Return True if atomic forces are also computed.
    """
    return self._atomic_forces

  @property
  def is_lj(self):
    """
    Return True if this transformer is used for LJ systems.
    """
    return self._lj

  @property
  def cutoff(self):
    """
    Return the cutoff.
    """
    return self._cutoff

  def accept_species(self, species):
    """
    Return True if the given species can be handled.
    
    Args:
      species: a `List[str]` as the ordered species of a molecule.
    
    Returns:
      accepted: True if the given species can be handled by this transformer.
    
    """
    counter = Counter(species)
    return all(counter[e] <= self._max_occurs.get(e, 0) for e in counter)

  def _get_transformer(self, species):
    """
    Return the `Transformer` for the given list of species.
    
    Args:
      species: a `List[str]` as the atomic species.

    Returns:
      clf: a `Transformer`.

    """
    species = list(species) + [GHOST] * self._num_ghosts
    formula = get_formula(species)
    clf = self._transformers.get(
      formula, Transformer(species=species,
                           k_max=self._k_max,
                           kbody_terms=self._kbody_terms,
                           split_dims=self._split_dims,
                           norm=self._norm,
                           norm_order=self._norm_order,
                           periodic=self._periodic,
                           atomic_forces=self._atomic_forces,
                           lj=self._lj,
                           cutoff=self._cutoff)
    )
    self._transformers[formula] = clf
    return clf

  def transform_trajectory(self, trajectory):
    """
    Transform the given trajectory (a list of `ase.Atoms` with the same chemical
    symbols or an `ase.io.TrajectoryReader`) to input features.

    Args:
      trajectory: a `list` of `ase.Atoms` or a `ase.io.TrajectoryReader`. All
        objects should have the same chemical symbols.

    Returns:
      sample: a `KcnnSample` object.

    """
    ntotal = len(trajectory)
    assert ntotal > 0

    species = trajectory[0].get_chemical_symbols()
    if not self.accept_species(species):
      raise ValueError(
        "This transformer does not support {}!".format(get_formula(species)))

    clf = self._get_transformer(species)
    split_dims = np.asarray(clf.split_dims)
    nrows, ncols = clf.shape
    features = np.zeros((ntotal, nrows, ncols), dtype=np.float32)

    if self._atomic_forces:
      coef = np.zeros((ntotal, nrows, ncols * 6), dtype=np.float32)
      num_force_components = 3 * len(species)
      num_entries = nrows * ncols * 6 // num_force_components
      indexing = np.zeros((ntotal, num_force_components, num_entries),
                          dtype=np.float32)
    else:
      coef = None
      indexing = None

    occurs = np.zeros((ntotal, self._num_atom_types), dtype=np.float32)
    for specie, times in Counter(species).items():
      loc = self._atom_types.index(specie)
      if loc < 0 or loc >= self._num_atom_types:
        raise ValueError("The loc of {:s} is {:d}!".format(specie, loc))
      occurs[:, loc] = float(times)

    weights = np.zeros((ntotal, nrows), dtype=np.float32)
    compress_stats = {}

    for i, atoms in enumerate(trajectory):
      _, coef_, indexing_ = clf.transform(atoms, features=features[i])
      if self._atomic_forces:
        # TODO: fix the bug
        coef[i] = coef_
        indexing[i] = indexing_
      if self._cutoff is None:
        weights[i] = clf.binary_weights
      else:
        weights[i], stats = clf.compress(features[i])
        for k, v in stats.items():
          compress_stats[k] = max(compress_stats.get(k, 0), v)

    return KcnnSample(features=features,
                      split_dims=split_dims,
                      binary_weights=weights,
                      occurs=occurs,
                      coefficients=coef,
                      indexing=indexing,
                      compress_stats=compress_stats)

  def transform(self, atoms):
    """
    Transform a single `ase.Atoms` object to input features.

    Args:
      atoms: an `ase.Atoms` object as the target to transform.

    Returns:
      sample: a `KcnnSample` object.
    
    Raises:
      ValueError: if the `species` is not supported by this transformer.

    """
    species = atoms.get_chemical_symbols()
    if not self.accept_species(species):
      raise ValueError(
        "This transformer does not support {}!".format(get_formula(species)))

    clf = self._get_transformer(species)
    split_dims = np.asarray(clf.split_dims)
    features, coef, indexing = clf.transform(atoms)
    occurs = np.zeros((1, self._num_atom_types), dtype=np.float32)
    for specie, times in Counter(species).items():
      loc = self._atom_types.index(specie)
      if loc < 0:
        raise ValueError("The loc of %s is -1!" % specie)
      occurs[0, loc] = float(times)
    if self._cutoff is not None:
      weights, compress_stats = clf.compress(features)
    else:
      weights = np.array(clf.binary_weights)
      compress_stats = {}
    return KcnnSample(features=features,
                      split_dims=split_dims,
                      binary_weights=weights,
                      occurs=occurs,
                      coefficients=coef,
                      indexing=indexing,
                      compress_stats=compress_stats)

  def compute_atomic_energies(self, species, y_kbody, y_atomic_1body):
    """
    Compute the atomic energies given predicted kbody contributions.
    
    Args:
      species: a `List[str]` as the ordered atomic species.
      y_kbody: a `float32` array of shape `[num_examples, N]` as the k-body
        contribs.
      y_atomic_1body: a `Dict[str, float]` as the one-body energy of each kind 
        of atom.

    Returns:
      y_atomic: a `float32` array of shape `[num_examples, num_atoms]`.

    """

    # Get the feature transformer
    clf = self._get_transformer(species)

    # Allocate arrays
    num_examples = y_kbody.shape[0]
    num_atoms = len(species)
    split_dims = clf.split_dims
    y_atomic = np.zeros((num_examples, num_atoms))

    # Setup the 1-body atomic energies.
    for i in range(num_atoms):
      y_atomic[:, i] = y_atomic_1body[species[i]]

    # Compute and add the higher order (k = 2, 3, ...) corrections.
    for step in range(num_examples):
      for i, kbody_term in enumerate(clf.kbody_terms):
        if kbody_term not in clf.kbody_selections:
          continue

        # Compute the real `k` for this k-body term by excluding ghost atoms.
        symbols = get_atoms_from_kbody_term(kbody_term)
        k = len(symbols) - symbols.count(GHOST)

        # For each k-atoms selection, its energy should contrib equally to the
        # selected atoms. In my paper the coef should be `1 / factorial(k)` but
        # since our k-atom selections are all unique so coef here is `1 / k`.
        coef = 1.0 / k

        # Locate the starting index
        istart = 0 if i == 0 else int(sum(split_dims[:i]))

        # Loop through all k-atoms selections
        for ki, indices in enumerate(clf.kbody_selections[kbody_term]):
          for ai in indices[:k]:
            y_atomic[step, ai] += y_kbody[step, istart + ki] * coef
    return y_atomic


class OneBodyCalculator:
  """
  A helper class to compute the initial one-body weights.
  """

  def __init__(self, atom_types, num_examples, algorithm='default', factor=1.0,
               include_perturbations=True):
    """
    Initialization method.

    Args:
      atom_types: a list of `str` as the types of atoms.
      num_examples: an `int` as the total number of examples.
      algorithm: a `str` as the algorithm to compute the one-body weights.
      factor: a `float` as the scaling factor as the one-body weights.
      include_perturbations: a `bool`. If True, the higher-order perturbations
        terms will be included in the coefficients matrix as well.

    """
    self.atom_types = atom_types
    self.num_atom_types = len(atom_types)
    if atom_types[-1] == GHOST:
      self.num_real_atom_types = self.num_atom_types - 1
    else:
      self.num_real_atom_types = self.num_atom_types
    self.minima = {}
    self.b = np.zeros((num_examples, ))
    self.algorithm = algorithm.lower()
    self.factor = factor
    self.include_perturbations = include_perturbations
    self.mp2 = self.num_real_atom_types
    self.mp3 = self.num_real_atom_types + 1
    if not include_perturbations:
      self.coef = np.zeros((num_examples, self.num_real_atom_types))
    else:
      self.coef = np.zeros((num_examples, self.num_real_atom_types + 2))

  def add(self, index, chemical_symbols, y_true):
    """
    Add an example.

    Args:
      index: an `int` as the index of this sample.
      chemical_symbols: a `list` of `str` as the chemical symbols of this
        example.
      y_true: a `float` as the total energy of this example.

    """
    counter = Counter(chemical_symbols)
    for loc, atom in enumerate(self.atom_types[:self.num_real_atom_types]):
      self.coef[index, loc] = counter[atom]
    if self.include_perturbations:
      self.coef[index, self.mp2] = comb(len(chemical_symbols), 2)
      self.coef[index, self.mp3] = comb(len(chemical_symbols), 3)
    self.b[index] = y_true
    sch = self.get_stoichiometry(self.coef[index, :self.num_real_atom_types])
    if sch not in self.minima or y_true < self.b[self.minima[sch]]:
      self.minima[sch] = index

  def compute(self):
    """
    Compute the one-body weights.
    """
    if self.algorithm == 'minimal':
      # Only select the values from the global minima.
      selected = np.ix_(list(self.minima.values()))
      coef = self.coef[selected]
      b = self.b[selected]
    else:
      coef = self.coef
      b = self.b
    # The size of `x` is always equal to `self.num_real_atom_types`. We may need
    # to pad an zero at the end.
    x = _compute_lr_weights(
      coef, b,
      num_real_atom_types=self.num_real_atom_types,
      factor=self.factor
    )
    x = np.resize(x, self.num_atom_types)
    x[self.num_real_atom_types:] = 0.0
    return x

  def get_stoichiometry(self, atoms_counts):
    """
    A helper function to get the stoichiometry of a structure.
    """
    return ";".join(["{},{}".format(self.atom_types[j], int(atoms_counts[j]))
                     for j in range(self.num_real_atom_types)])


class FixedLenMultiTransformer(MultiTransformer):
  """
  This is also a flexible transformer targeting on AxByCz ... molecular 
  compositions but the length of the transformed features are the same so that
  they can be used for training.
  """

  def __init__(self, max_occurs, periodic=False, k_max=3, norm='exp',
               norm_order=1, include_all_k=True, atomic_forces=False, lj=False,
               cutoff=None):
    """
    Initialization method. 
    
    Args:
      max_occurs: a `Dict[str, int]` as the maximum appearances for each kind of 
        atomic specie.
      periodic: a `bool` indicating whether this transformer is used for 
        periodic structures or not.
      k_max: a `int` as the maximum k for the many body expansion scheme.
      norm: a `str` specifying the normalization function to use. Defaults to
        `exp`, `lj`, `exp+g`, 'inv' and `morse` are also supported.
      norm_order: an `int` or a `float` as the order for normalizing interatomic
        distances.
      include_all_k: a `bool` indicating whether we shall include all k-body
        terms where k = 1, ..., k_max. If False, only k = 1 and k = k_max are
        considered.
      atomic_forces: a `bool` indicating whether the atomic forces derivation is
        enabled or not.
      lj: a `bool` indicating that this transformer targets on LJ systems.
      cutoff: a `float` as the cutoff.
    
    """
    super(FixedLenMultiTransformer, self).__init__(
      atom_types=list(max_occurs.keys()),
      k_max=k_max,
      max_occurs=max_occurs,
      norm=norm,
      norm_order=norm_order,
      include_all_k=include_all_k,
      periodic=periodic,
      atomic_forces=atomic_forces,
      lj=lj,
      cutoff=cutoff
    )
    self._split_dims = self._get_fixed_split_dims()
    self._total_dim = sum(self._split_dims)

  @property
  def shape(self):
    """
    Return the shape of input feature matrix of this transformer.
    """
    return self._total_dim, comb(self._k_max, 2, exact=True)

  @property
  def split_dims(self):
    """
    Return the fixed `split_dims` for all internal transformers.
    """
    return self._split_dims

  def _get_fixed_split_dims(self):
    """
    The `split_dims` of all `_Transformer` should be the same.
    """
    split_dims = []
    for kbody_term in self._kbody_terms:
      atoms = get_atoms_from_kbody_term(kbody_term)
      counter = Counter(atoms)
      dims = [comb(self._max_occurs[e], k, True) for e, k in counter.items()]
      split_dims.append(np.prod(dims))
    return [int(x) for x in split_dims]

  def _log_compression_results(self, results):
    """
    A helper function to log the compression result.
    """
    print("The soft compression algorithm is applied with cutoff = "
          "{:.2f}".format(self._cutoff))

    num_loss_total = 0
    num_total = self._total_dim
    for j, kbody_term in enumerate(self._kbody_terms):
      num_full = self._split_dims[j]
      num_kept = results.get(kbody_term, 0)
      num_loss = num_full - num_kept
      num_loss_total += num_loss
      print("{:<12s} : {:5d} / {:5d}, compression = {:.2f}%".format(
        kbody_term, num_loss, num_full, num_loss / num_full * 100))
    print("Final result : {:5d} / {:5d}, compression = {:.2f}%".format(
      num_loss_total, num_total, num_loss_total / num_total * 100))

  def _transform_and_save(self, filename, examples, num_examples, max_size,
                          loss_fn=None, verbose=True, one_body_kwargs=None):
    """
    Transform the given atomic coordinates to input features and save them to
    tfrecord files using `tf.TFRecordWriter`.

    Args:
      filename: a `str` as the file to save examples.
      examples: a iterator which iterates through all examples.
      num_examples: an `int` as the number of examples.
      max_size: an `int` as the maximum size of all structures. This determines
        the dimension of the forces.
      verbose: boolean indicating whether.
      loss_fn: a `Callable` for transforming the calculated raw loss.
      one_body_kwargs: a `dict` as the key-value args for computing initial
        one-body weights.

    Returns:
      weights: a `float32` array as the weights for linear fit of the energies.

    """

    def _identity(_):
      """
      An identity function which returns the input directly.
      """
      return 1.0

    # Setup the loss function.
    loss_fn = loss_fn or _identity

    # Setup the one-body weights calculator
    one_body = OneBodyCalculator(
      self._atom_types, num_examples, **(one_body_kwargs or {}))

    # Start the timer
    tic = time.time()

    with tf.python_io.TFRecordWriter(filename) as writer:
      if verbose:
        print("Start transforming {} ... ".format(filename))

      compress_stats = {}

      for i, atoms in enumerate(examples):

        species = atoms.get_chemical_symbols()
        y_true = atoms.get_total_energy()
        sample = self.transform(atoms)

        x = _bytes_feature(sample.features.tostring())
        y = _bytes_feature(np.atleast_2d(-y_true).tostring())
        z = _bytes_feature(sample.occurs.tostring())
        w = _bytes_feature(sample.binary_weights.tostring())
        y_weight = _float_feature(loss_fn(y_true))

        if not self._atomic_forces:
          example = Example(
            features=Features(feature={'energy': y, 'features': x, 'occurs': z,
                                       'weights': w, 'loss_weight': y_weight}))
        else:
          # Pad zeros to the forces so that all forces of this dataset have the
          # same dimension.
          forces = atoms.get_forces()
          pad = max_size - len(forces)
          if pad > 0:
            forces = np.pad(forces, ((0, pad), (0, 0)), mode='constant')
          forces = _bytes_feature(forces.flatten().tostring())
          coef = _bytes_feature(sample.coefficients.tostring())
          indexing = _bytes_feature(sample.indexing.tostring())
          example = Example(
            features=Features(feature={'energy': y, 'features': x, 'occurs': z,
                                       'weights': w, 'loss_weight': y_weight,
                                       'indexing': indexing, 'coef': coef,
                                       'forces': forces}))
        writer.write(example.SerializeToString())

        # Add this example to the one-body database
        one_body.add(i, species, y_true)

        # Save the compress stats for this example
        for k, v in sample.compress_stats.items():
          compress_stats[k] = max(compress_stats.get(k, 0), v)

        if verbose and (i + 1) % 100 == 0:
          sys.stdout.write("\rProgress: {:7d} / {:7d} | Speed = {:6.1f}".format(
            i + 1, num_examples, (i + 1) / (time.time() - tic)))

      if verbose:
        print("")
        print("Transforming {} finished!".format(filename))

        if self._cutoff is not None:
          self._log_compression_results(compress_stats)

      return one_body.compute()

  def _save_auxiliary_for_file(self, filename, max_size, lookup_indices=None,
                               initial_1body_weights=None):
    """
    Save auxiliary data for the given dataset.

    Args:
      filename: a `str` as the tfrecords file.
      max_size: an `int` as maximum size of all structures. This determines the
        dimension of the forces.
      initial_1body_weights: a `float32` array of shape `[num_atom_types, ]` as
        the initial weights for the one-body convolution kernels.
      lookup_indices: a `List[int]` as the indices of each given example.

    """
    if lookup_indices is not None:
      lookup_indices = list(lookup_indices)
    else:
      lookup_indices = []

    if initial_1body_weights is not None:
      initial_1body_weights = list(initial_1body_weights)
    else:
      initial_1body_weights = []

    max_occurs = {atom: times for atom, times in self._max_occurs.items()
                  if times < self._k_max}
    num_entries = _get_num_force_entries(max_size, self._k_max)

    auxiliary_properties = {
      "kbody_terms": self._kbody_terms,
      "split_dims": self._split_dims,
      "shape": self.shape,
      "lookup_indices": list([int(i) for i in lookup_indices]),
      "atom_types": self._atom_types,
      "num_atom_types": self._num_atom_types,
      "species": self._species,
      "include_all_k": self._include_all_k,
      "periodic": self._periodic,
      "k_max": self._k_max,
      "max_occurs": max_occurs,
      "norm": self._norm,
      "norm_order": self._norm_order,
      "initial_one_body_weights": initial_1body_weights,
      "atomic_forces_enabled": self._atomic_forces,
      "indexing_shape": [max_size * 3, num_entries],
      "lj": self._lj,
      "cutoff": self._cutoff
    }

    with open(join(dirname(filename),
                   "{}.json".format(splitext(basename(filename))[0])),
              "w+") as fp:
      json.dump(auxiliary_properties, fp=fp, indent=2)

  def transform_and_save(self, database, train_file=None, test_file=None,
                         loss_fn=None, verbose=True, one_body_kwargs=None):
    """
    Transform coordinates to input features and save them to tfrecord files
    using `tf.TFRecordWriter`.

    Args:
      database: a `Database` as the parsed results from a xyz file.
      train_file: a `str` as the file for saving training data or None to skip.
      test_file: a `str` as the file for saving testing data or None to skip.
      verbose: a `bool` indicating whether logging the transformation progress.
      loss_fn: a `Callable` for computing the exponential scaled RMSE loss.
      one_body_kwargs: a `dict` as the configs for the initial one-body weigts
        calculator.

    """
    #fix a bug by Jinzhe Zeng
    max_size = len(self._species) - self._num_ghosts

    if test_file:
      examples = database.examples(mode=tf.estimator.ModeKeys.EVAL)
      id_list = database.ids_of_testing_examples
      num_examples = len(id_list)
      if num_examples > 0:
        self._transform_and_save(
          test_file,
          examples,
          num_examples,
          max_size,
          loss_fn=loss_fn,
          verbose=verbose
        )
        self._save_auxiliary_for_file(
          test_file,
          max_size=max_size,
          lookup_indices=id_list
        )

    if train_file:
      examples = database.examples(mode=tf.estimator.ModeKeys.TRAIN)
      id_list = database.ids_of_training_examples
      num_examples = len(id_list)
      if num_examples > 0:
        weights = self._transform_and_save(
          train_file,
          examples,
          num_examples,
          max_size,
          loss_fn=loss_fn,
          verbose=verbose,
          one_body_kwargs=one_body_kwargs or {}
        )
        self._save_auxiliary_for_file(
          train_file,
          max_size=max_size,
          initial_1body_weights=weights,
          lookup_indices=id_list
        )



# coding=utf-8
"""
This module offers auxiliary functions for add summaries.
"""
from __future__ import print_function, absolute_import

import tensorflow as tf

__author__ = 'Xin Chen'
__email__ = 'Bismarrck@me.com'


def add_total_norm_summaries(grads_and_vars, collection,
                             only_summary_total=True):
  """
  Add summaries for the 2-norms of the gradients.

  Args:
    grads_and_vars: a list of (gradient, variable) returned by an optimizer.
    collection: a `str` as the collection to add the computed norms.
    only_summary_total: a `bool` indicating whether we should summarize the
      individual norms or not.

  Returns:
    list_of_ops: a list of added summary tensors.

  """
  list_of_ops = []

  for grad, var in grads_and_vars:
    if grad is not None:
      norm = tf.norm(grad, name=var.op.name + "/norm")
      tf.add_to_collection(collection, norm)
      with tf.name_scope("gradients/{}/".format(collection)):
        list_of_ops.append(tf.summary.histogram(var.op.name + "/hist", grad))
      if not only_summary_total:
        with tf.name_scope("gradients/{}/".format(collection)):
          list_of_ops.append(tf.summary.scalar(var.op.name + "/norm", norm))

  with tf.name_scope("total_norm/"):
    total_norm = tf.add_n(tf.get_collection(collection))
    list_of_ops.append(tf.summary.scalar(collection, total_norm))
  return list_of_ops


def add_variable_summaries():
  """
  Add variable summaries.

  Returns:
    list_of_ops: a list of added summary tensors.

  """
  list_of_ops = []

  with tf.name_scope("variables"):
    for var in tf.trainable_variables():
      list_of_ops.append(tf.summary.histogram(var.op.name, var))
      vsum = tf.reduce_sum(tf.abs(var, name="absolute"), name="vsum")

      if not var.op.name.startswith('kCON/one-body'):
        tf.add_to_collection('k_sum', vsum)
      else:
        tf.add_to_collection('1_sum', vsum)

    list_of_ops.append(tf.summary.scalar(
      'kbody', tf.add_n(tf.get_collection('k_sum'), name='vars_ksum')))
    list_of_ops.append(tf.summary.scalar(
      '1body', tf.add_n(tf.get_collection('1_sum'), name='vars_1sum')))

  return list_of_ops



import logging
import threading
import tempfile
import os

import pkg_resources
import pytest

from kconmd import kconMD
from kconmd.server import kconMD_client, kconMD_server


class Test_all:
    @pytest.fixture(scope="class")
    def kconmd(self):
        folder = tempfile.mkdtemp(prefix='testfiles-', dir='.')
        logging.info(f'Folder: {folder}:')
        os.chdir(folder)
        xyzfilename = 'test.xyz'
        pbfilename = 'test.pb'
        forcefilename = 'force'

        for filename in (xyzfilename, pbfilename):
            with open(filename, 'wb') as f:
                f.write(pkg_resources.resource_string(__name__, filename))
        return kconMD(pbfilename, xyzfilename, forcefilename)

    def test_kconmd(self, kconmd):
        kconmd.printforce()
        self._printresult(kconmd)

    def test_server(self, kconmd):
        server = kconMD_server(kconmd)
        cilent = kconMD_client()
        t = threading.Thread(target=server.response, name='server')
        t.start()
        cilent.printforce()
        cilent.exitserver()
        self._printresult(kconmd)
        t.join()

    @classmethod
    def _printresult(cls, kconmd):
        with open(kconmd.outputfilename) as f:
            logging.info("Force:")
            print(f.read())
